------------------------
--Map location sharing--
------------------------
--This module shows map markers for friends/guild/all depending on your settings in Multimod options.
--Without the need to be in a party.

local playerFaction = UnitFactionGroup("player")
mmNameAndServer = UnitName("player") .. "-" .. _G.GetRealmName()
--NOTE: This cache only stores people on friends/bnet list with the addon, no need for guild/all.
local mmMapCache = {}

--Send one map update after joining the addon channel.
local mapUpdateOnce = CreateFrame("Frame")
mapUpdateOnce:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")
mapUpdateOnce:SetScript('OnEvent', function(self, event, type, b, c, d, e, f, g, h, channelname)
	if (type == "YOU_CHANGED" and channelname == "mmcomms") then
		sendMapUpdate(true)
		mapUpdateOnce:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE")
		mapUpdateOnce = nil
	end
end)

--Coords received from someone.
--The arg "playerOrID" can also be a bnet presenceID.
function mapCommReceived(playerOrID, string, distribution)
	local data = explode(",", string)
    local player = tostring(data[1])
    local class = tostring(data[2])
    local groupCheck = true
	--UnitInParty() doesn't return true if you include realm and the player is on your server so we have to strip it.
	local realm = string.sub(player, string.find(player, "-") + 1)
	if (realm == Multimod.homeServer or _G.GetRealmName()) then
		groupCheck = string.sub(player, 1, string.find(player, "-") - 1)
	else
		groupCheck = player
	end
	if (player == mmNameAndServer or UnitInRaid(groupCheck) or UnitInParty(groupCheck)) then
		return
	end
    if (string and string ~= nil and string ~= '') then
    	if  (class == "stop") then
    		--When someone disables broadcasting in options they send a stop command to remove the stale map marker.
    		removeMapMarker(player)
    		return
    	end
    	--Register addon users.
    	if (class == "addoncheck") then
    		addMmMapCache(player)
    		if (distribution == "BNET") then
    			BNSendGameData(playerOrID, Multimod.commPrefix, "mmmap " .. mmNameAndServer .. ",addonreceived")
    		else
    			Multimod:sendAddonChannel("WHISPER", "mmmap " .. mmNameAndServer .. ",addonreceived", player)
    		end
    		return
    	end
    	if (class == "addonreceived") then
    		addMmMapCache(player)
 			sendMapUpdate(true)
    		return
    	end
    	if (Multimod.db.global.map_disable_broadcasting) then
			return
		end
    	local x = tonumber(data[3])
    	local y = tonumber(data[4])
    	local instance = tonumber(data[5])
    	local level = tonumber(data[6])
		if (distribution == "CHANNEL" and Multimod.db.global.map_all) then
			updateMarker(player, class, x, y, instance, level)
		elseif (distribution == "GUILD" and Multimod.db.global.map_guild) then
			updateMarker(player, class, x, y, instance, level)
		elseif (distribution == "BNET" and Multimod.db.global.map_bnetfriends) then
			updateMarker(player, class, x, y, instance, level)
		elseif (distribution == "WHISPER" and Multimod.db.global.map_friends) then
			updateMarker(player, class, x, y, instance, level)
		end
   end
end

--Create dynamically named map markers for each player.
--Markers are stored as NameRealm without the "-".
local markerNames = {}
local mapMarkers = {}
function createMapMarker(varFriendlyName, class, fullName)
	--fullName is just used for the hover over map tolltip.
	local splitFullName = explode("-", fullName)
	local nameOnly = splitFullName[1]
	local realmOnly = splitFullName[2]
	mapMarkers[varFriendlyName] = {}
	local iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinblackborder\\" .. class
	if (Multimod.db.global.map_marker_icon == 1) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Solidcolor\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 2) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinblackborder\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 3) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickblackborder\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 4) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickestblackborder\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 5) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinorangeborder\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 6) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickorangeborder\\" .. class
	elseif (Multimod.db.global.map_marker_icon == 7) then
		iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickestorangeborder\\" .. class
	end
	--Map marker frame.
	local obj = CreateFrame("Button", "mmMapMarker" .. varFriendlyName, WorldMapButton)
	obj.bg = obj:CreateTexture(nil,	"ARTWORK")
	obj.bg:SetTexture(iconLocation)
	obj.bg:SetAllPoints(obj)
	obj.texture = obj.bg
	obj:SetWidth(15)
	obj:SetHeight(15)
	obj.class = class
	obj.varFriendlyName = varFriendlyName
	--Map marker tooltip showing players name on hover.
	obj.Tooltip = CreateFrame("FRAME", "mmMapMarkerTooltip" .. varFriendlyName, obj, "ThinGoldEdgeTemplate")
	obj.Tooltip:SetPoint("Top", 0, 25)
	obj.Tooltip.fs = obj.Tooltip:CreateFontString("mapMarkerTtFS", "ARTWORK")
	obj.Tooltip.fs:SetFont("Fonts\\ARIALN.ttf", 20)
	if (realmOnly == _G.GetRealmName()) then
		obj.Tooltip.fs:SetText("|CffDEDE42" .. nameOnly)
	else
		obj.Tooltip.fs:SetText("|CffDEDE42" .. fullName)
	end
	obj.Tooltip.fs:SetPoint("CENTER", 0.5, 0)
	local fswidth = obj.Tooltip.fs:GetStringWidth()
	local fsheight = obj.Tooltip.fs:GetStringHeight()
	obj.Tooltip:SetWidth(fswidth + 10)
	obj.Tooltip:SetHeight(fsheight + 6)
	obj.Tooltip:SetFrameLevel(1000)
    obj.Tooltip:SetFrameStrata("TOOLTIP")
	obj.Tooltip:Hide()
	function obj.show()
		mapMarkers[varFriendlyName]['worldmap'].Tooltip:Show()
	end
	function obj.hide()
		mapMarkers[varFriendlyName]['worldmap'].Tooltip:Hide()
	end
	obj:SetScript("OnEnter", obj.show)
	obj:SetScript("OnLeave", obj.hide)
	mapMarkers[varFriendlyName]['worldmap'] = obj
	--mmMiniMarkerNovaoneFrostmourne:SetFrameStrata("BACKGROUND")
	--mmMiniMarkerNovaoneFrostmourne:SetFrameLevel(3)
	--mmMiniMarkerNovaoneFrostmourne.texture:SetDrawLayer("Background", -8)
	--Minimap marker frame.
	local obj = CreateFrame("FRAME", "mmMiniMarker" .. varFriendlyName)
	obj.bg = obj:CreateTexture(nil, "BACKGROUND", nil, -8)
	obj.bg:SetTexture(iconLocation)
	obj.bg:SetAllPoints(obj)
	obj.texture = obj.bg
	obj:SetWidth(12)
	obj:SetHeight(12)
	obj.class = class
	--Minimap marker tooltip showing player name on hover.
	obj.miniMarkerTooltip = CreateFrame("FRAME", "mmMiniMarkerTooltip" .. varFriendlyName, obj, "ThinGoldEdgeTemplate")
	obj.miniMarkerTooltip:SetPoint("Top", 0, 25)
	obj.miniMarkerTooltip.fs = obj.miniMarkerTooltip:CreateFontString("miniMarkerTtFS", "ARTWORK")
	obj.miniMarkerTooltip.fs:SetFont("Fonts\\ARIALN.ttf", 14)
	if (realmOnly == _G.GetRealmName()) then
		obj.miniMarkerTooltip.fs:SetText("|CffDEDE42" .. nameOnly)
	else
		obj.miniMarkerTooltip.fs:SetText("|CffDEDE42" .. fullName)
	end
	obj.miniMarkerTooltip.fs:SetPoint("CENTER", 0.5, 0)
	local fswidth = obj.miniMarkerTooltip.fs:GetStringWidth()
	local fsheight = obj.miniMarkerTooltip.fs:GetStringHeight()
	obj.miniMarkerTooltip:SetWidth(fswidth + 10)
	obj.miniMarkerTooltip:SetHeight(fsheight + 6)
	obj.miniMarkerTooltip:SetFrameLevel(1000)
    obj.miniMarkerTooltip:SetFrameStrata("TOOLTIP")
	obj.miniMarkerTooltip:Hide()
	function obj.show()
		mapMarkers[varFriendlyName]['minimap'].miniMarkerTooltip:Show()
	end
	function obj.hide()
		mapMarkers[varFriendlyName]['minimap'].miniMarkerTooltip:Hide()
	end
	obj:SetScript("OnEnter", obj.show)
	obj:SetScript("OnLeave", obj.hide)
	mapMarkers[varFriendlyName]['minimap'] = obj
	mapMarkers[varFriendlyName]['fullname'] = fullName
	
	--Keep a local array of marker names names.
	table.insert(markerNames, "mmMapMarker" .. varFriendlyName)
	table.insert(markerNames, "mmMiniMarker" .. varFriendlyName)
end

--If someone leaves addon or goes offline on bnet/friends then remove marker.
function removeMapMarker(player)
	local player = string.gsub(player, "-", "")
	if (_G["mmMapMarker" .. player]) then
		mmDragonPins:RemoveAllWorldMapIcons("mmwm" .. player)
		mmDragonPins:RemoveAllMinimapIcons("mmmm" .. player)
	end
end

--Called when someone joins/leaves group.
Multimod.doMapRosterUpdate = true
local f = CreateFrame("Frame")
f:RegisterEvent("GROUP_ROSTER_UPDATE")
f:SetScript('OnEvent', function(self, event, ...)
	if (Multimod.doMapRosterUpdate) then
		--Add a throddle so as not to spam func when joining large groups.
		Multimod.doMapRosterUpdate = false
		C_Timer.After(2, function()
			removeAllMarkers(true)
		end)
	end
end)

--Remove all markers, if groupOnly is true then only remove player markers in your group, checked on GROUP_ROSTER_UPDATE. 
function removeAllMarkers(groupOnly)
	Multimod.doMapRosterUpdate = true
	for i, v in ipairs(markerNames) do
		if (_G[v]) then
			if (string.find(v, "mmMapMarker")) then
				player = string.match(v, 'mmMapMarker(.+)')
				--Worldmap marker.
				if (groupOnly and mapMarkers[player]) then
					local realm = string.sub(mapMarkers[player]['fullname'], string.find(mapMarkers[player]['fullname'], "-") + 1)
					if (realm == Multimod.homeServer or _G.GetRealmName()) then
						groupCheck = string.sub(player, 1, string.find(mapMarkers[player]['fullname'], "-") - 1)
					else
						groupCheck = player
					end
					if (UnitInParty(groupCheck) or UnitInRaid(groupCheck)) then
						mmDragonPins:RemoveAllWorldMapIcons("mmwm" .. player)
					end
				else
					mmDragonPins:RemoveAllWorldMapIcons("mmwm" .. player)
				end
			end
			if (string.find(v, "mmMiniMarker")) then
				player = string.match(v, 'mmMiniMarker(.+)')
				--Minimap marker.
				if (groupOnly and mapMarkers[player]) then
					local realm = string.sub(mapMarkers[player]['fullname'], string.find(mapMarkers[player]['fullname'], "-") + 1)
					if (realm == Multimod.homeServer or _G.GetRealmName()) then
						groupCheck = string.sub(player, 1, string.find(mapMarkers[player]['fullname'], "-") - 1)
					else
						groupCheck = player
					end
					if (UnitInParty(groupCheck) or UnitInRaid(groupCheck)) then
						mmDragonPins:RemoveAllMinimapIcons("mmmm" .. player)
					end
				else
					mmDragonPins:RemoveAllMinimapIcons("mmmm" .. player)
				end
			end
		end
	end
end

--When someone changes texture options.
function reloadMapMarkers()
	for i, v in ipairs(markerNames) do
		if (Multimod.db.global.map_marker_icon == 1) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Solidcolor\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 2) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinblackborder\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 3) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickblackborder\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 4) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickestblackborder\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 5) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinorangeborder\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 6) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickorangeborder\\" .. _G[v].class
		elseif (Multimod.db.global.map_marker_icon == 7) then
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thickestorangeborder\\" .. _G[v].class
		else
			iconLocation = "Interface\\AddOns\\Multimod\\Media\\Mapicons\\Thinblackborder\\" .. _G[v].class
		end
		_G[v].bg:SetTexture(iconLocation)
	end
end

--Fires when someone disables broadcasting map coords in options.
function stopBroadcasting()
	Multimod:sendAddonChannel("CHANNEL", "mmmap " .. mmNameAndServer .. ",stop")
	sendBnetFriends("mmmap " .. mmNameAndServer .. ",stop")
	sendFriends("mmmap " .. mmNameAndServer .. ",stop")
end

--Fires when someone disables the public addon channel broadcasting (but may still have guild/friends on).
function stopPublic()
	Multimod:sendAddonChannel("CHANNEL", "mmmap " .. mmNameAndServer .. ",stop")
end

--Fires when someone disables the guild addon channel broadcasting.
function stopGuild()
	if (IsInGuild()) then
		Multimod:sendAddonChannel("GUILD", "mmmap " .. mmNameAndServer .. ",stop")
	end
end

--Add a marker to the map.
function updateMarker(player, class, x, y, instance, level)
	if (not level or level == nil or level == '') then
		level = 0
	end
	local varFriendlyName = string.gsub(player, "-", "")
	if (not _G["mmMapMarker" .. varFriendlyName] or not _G["mmMiniMarker" .. varFriendlyName]) then
		createMapMarker(varFriendlyName, class, player)
	end
	if (x ~= nil and y ~= nil and instance ~= nil) then
		mmDragonPins:AddWorldMapIconMap("mmwm" .. varFriendlyName, _G["mmMapMarker" .. varFriendlyName], instance, x, y, HBD_PINS_WORLDMAP_SHOW_WORLDo)
	end
	local xx, yy, ii = mmDragonLib:GetWorldCoordinatesFromZone(x, y, instance, level)
	if (xx ~= nil and yy ~= nil and ii ~= nil) then
		local oX, oY, oZone, oLevel = mmDragonLib:GetPlayerZonePosition()
		local distance, deltaX, deltaY = mmDragonLib:GetZoneDistance(oZone, oLevel, oX, oY, instance, level, x, y)
		--Remove minimap marker from edge of minimap if set in options.
		if ((distance ~= nil and Multimod.db.global.map_minimarker_edge and distance > Multimod.db.global.map_minimarker_edge_distance) or
				Multimod.db.global.map_minimarker_edge_always) then
			mmDragonPins:AddMinimapIconWorld("mmmm" .. varFriendlyName, _G["mmMiniMarker" .. varFriendlyName], ii, xx, yy)
		else
			mmDragonPins:AddMinimapIconWorld("mmmm" .. varFriendlyName, _G["mmMiniMarker" .. varFriendlyName], ii, xx, yy, true)
		end
	end
end

--Broadcast coords.
local total = 0
local lastPosition = "0,0,0"
local _, sendClass = UnitClass("player");
local function onUpdate(self,elapsed)
    total = total + elapsed
    if total >= 2 then
    	total = 0
        sendMapUpdate()
    end
end
local f = CreateFrame("frame")
f:SetScript("OnUpdate", onUpdate)

--Update every 2 seconds.
function sendMapUpdate(sendOnce)
	if (sendOnce) then
		--Make sure next cycle it sends even if at same spot.
		lastPosition = nil
		return
	end
	if (Multimod.db.global.map_disable_broadcasting) then
		return
	end
	local x, y, instance, level = mmDragonLib:GetPlayerZonePosition()
	if (not x or not y or not instance) then
		return
	end
	local currentPosition = x .. "," .. y .. "," .. instance .. "," .. level
	if (lastPosition ~= currentPosition) then
		positionString = mmNameAndServer .. "," .. sendClass .. "," .. currentPosition
		if (x ~= nil and y ~= nil) then
			if (Multimod.db.global.map_all) then
				Multimod:sendAddonChannel("CHANNEL", "mmmap " .. positionString)
				if (IsInGuild()) then
					Multimod:sendAddonChannel("GUILD", "mmmap " .. positionString)
				end
				sendBnetFriends("mmmap " .. positionString)
				sendFriends("mmmap " .. positionString)
			else
				if (IsInGuild() and Multimod.db.global.map_guild) then
					Multimod:sendAddonChannel("GUILD", "mmmap " .. positionString)
				end
				if (Multimod.db.global.map_bnetfriends) then
					sendBnetFriends("mmmap " .. positionString)
				end
				if (Multimod.db.global.map_friends) then
					sendFriends("mmmap " .. positionString)
				end
			end
			lastPosition = currentPosition
		else
			--If no coords found remove your marker.
			Multimod:sendAddonChannel("CHANNEL", "mmmap " .. UnitName("player") .. "," .. sendClass .. ",0,0,0")
		end
	end
end

--Send coords to bnet friends of same faction.
function sendBnetFriends(string)
	--Keep a temp cache of bnet ID's we send to during this func so we don't spam multiboxers more than once.
	if (not BNFeaturesEnabledAndConnected()) then
		return;
	end
	local sentCache = {}
	local totalBNFriends, onlineBNFriends = BNGetNumFriends()
	if (onlineBNFriends ~= nil) then
		for friendIndex = 1, onlineBNFriends do
			if ((BNGetFriendGameAccountInfo(friendIndex, 1) ~= nil) and (BNGetFriendGameAccountInfo(friendIndex, 1) ~= '')) then
				local i = 1
				_, toonName, _, realmName, _, faction = BNGetFriendGameAccountInfo(friendIndex, i)
				local _, _, _, _, _, presenceID = BNGetFriendInfo(friendIndex)
    			while toonName ~= nil and toonName ~= '' do
    				factionArray = explode("\n", faction)
    				faction = factionArray[1]
    				toonName = toonName .. "-" .. realmName
    				if (faction == playerFaction and mmMapCache[toonName] and not sentCache[presenceID]) then
    					sentCache[presenceID] = true
						BNSendGameData(presenceID, Multimod.commPrefix, string)
					end
      				i = i + 1
      				_, toonName = BNGetFriendGameAccountInfo(friendIndex, i)
      			end
    		end
		end
	end
end

--Send coords to people on normal ingame friends list.
function sendFriends(string)
	local totalFriends, onlineFriends = GetNumFriends()
	if (onlineFriends ~= nil) then
		for i = 1, onlineFriends do
			local name, level, class, area, connected, status, note = GetFriendInfo(i);
			if (string.match(name, "%-") == nil) then
				name = name .. "-" .. _G.GetRealmName()
			end
			if (name ~= nill and name ~= '' and mmMapCache[name] == true) then
				Multimod:sendAddonChannel("WHISPER", string, name)
			end
		end
	end
end

--Remove mapmarker when a friend logs off.
--(Friend logoffs are done in Tools.lua)

--Remove mapmarker when someone leaves addon channel (assume logoff).
local f = CreateFrame("Frame")
f:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE")
f:SetScript('OnEvent', function(self, event, a, name, b, c, d, e, f, g, channelname)
	--CHAT_MSG_CHANNEL_LEAVE always gives full Name-Realm
	if (channelname == "mmcomms") then
		local data = explode("-", name)
		local splitname = data[1]
		local realm = data[2]
		if (realm == Multimod.homeServer) then
			removeMapMarker(name)
		else
			removeMapMarker(name)
		end
	end
end)

--NOTE: This cache only stores people on friends/bnet list with the addon, no need for guild/all.
--Register friends with the addon so we don't send coords to friends without it.
--When friends log on a check is sent from friendCache() in Tools.lua. 
--Run at logon.
function createMapCache(doDelay)
	local sentCache = {}
	for name, v in pairs(onlineFriendCache) do
		if (v["type"] == "bnet" and BNFeaturesEnabledAndConnected()) then
			--Ace3 doesn't support sending bnet msgs yet so this is done without using the lib.
			local totalBNFriends, onlineBNFriends = BNGetNumFriends()
			if (onlineBNFriends ~= nil) then
				for friendIndex = 1, onlineBNFriends do
					local _, _, _, _, _, presenceID = BNGetFriendInfo(friendIndex)
					if ((BNGetFriendGameAccountInfo(friendIndex, 1) ~= nil) and (BNGetFriendGameAccountInfo(friendIndex, 1) ~= '')) then
						local i = 1
						_, toonName, _, realmName, _, faction = BNGetFriendGameAccountInfo(friendIndex, i)
    					while toonName ~= nil and toonName ~= '' do
    						if (toonName .. "-" .. realmName == name and not mmMapCache[toonName .. "-" .. realmName]
    								and not sentCache[presenceID]) then
    							sentCache[presenceID] = true
    							BNSendGameData(presenceID, Multimod.commPrefix, "mmmap " .. mmNameAndServer .. ",addoncheck")
    						end
      						i = i + 1
      						_, toonName = BNGetFriendGameAccountInfo(friendIndex, i)
      					end
    				end
				end
			end
		elseif (not mmMapCache[name]) then
			Multimod:sendAddonChannel("WHISPER", "mmmap " .. mmNameAndServer .. ",addoncheck", name)
		end
	end
	--Delay for a couple of seconds on a new bnet friend being added or removed, it needs time to update.
	if (doDelay) then
		C_Timer.After(2, function()
			sendMapUpdate(true)
		end)
	else
		sendMapUpdate(true)
	end
end

--Add to local addon user cache.
function addMmMapCache(player)
	if (string.match(player, "%-") == nil) then
		player = player .. "-" .. _G.GetRealmName()
	end
	mmMapCache[player] = true
end

--Remove from local addon user cache.
function remMmMapCache(player)
	if (string.match(player, "%-") == nil) then
		player = player .. "-" .. _G.GetRealmName()
	end
	mmMapCache[player] = nil
end

function printCache()
	for name, v in pairs(mmMapCache) do
		if (mmMapCache[name] == true) then
			print("Cache: " .. name)
		end
	end
end

--Set Map options at logon.
function Multimod:updateMapOptions(value)
	if (Multimod.db.global.map_bigger_arrow) then
		Minimap:SetPlayerTexture("Interface\\Minimap\\Vehicle-SilvershardMines-Arrow")
	elseif (value == "normalarrow") then
		Minimap:SetPlayerTexture("Interface\\Minimap\\MinimapArrow")
	end
end