--------------
---MultiMod---
--------------
--Contains commands and features to make multibox (and solo play) life easier.
--This is just a collection of things I've created along the way to help with certain things I do ingame.
--Made by Venomisto-Frostmourne OCE/US
--https://wow.curseforge.com/members/Venomisto68047

Multimod = LibStub("AceAddon-3.0"):NewAddon("Multimod", "AceComm-3.0")
mmDragonLib = LibStub("HereBeDragons-2.0")
mmDragonPins = LibStub("HereBeDragons-Pins-2.0")
mmLibRealmInfo = LibStub("LibRealmInfo")
mmAceGUI = LibStub("AceGUI-3.0")
Multimod.commPrefix = "mmData"
Multimod.addonChannel = "mmcomms"

--Addon loaded.
function Multimod:OnInitialize()
	--db is for saving addon options.
	--data is for saving internal data, keystone lists etc.
    self.db = LibStub("AceDB-3.0"):New("MultimodOptions", option_defaults, "Default")
    self.data = LibStub("AceDB-3.0"):New("MultimodData", data_defaults, "Default")
    --mmoptions.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    LibStub("AceConfig-3.0"):RegisterOptionsTable("Multimod", mmoptions)
	self.MultimodOptions = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Multimod", "Multimod")
	Multimod:RegisterComm(Multimod.commPrefix)
	Multimod:setHomeServer()
	Multimod:wipeOldKeystoneData()
	Multimod:talkingHeadLoad()
	Multimod:setMoveableInterfaceOptions()
	Multimod:updateChatOptions()
	Multimod:loadChatColor()
	--Multimod:questTrackerLoad()
	--Multimod:setQuestTracker()
end

--Set home server var.
function Multimod:setHomeServer()
	Multimod.homeServer = _G.GetRealmName()
	Multimod.nameAndServer = UnitName("player") .. "_" .. _G.GetRealmName() --This uses underscore so it can be used as a var.
	--I've been told you can't get realm name at load time even though it works for me when I try.
	--This is just a failsafe incase it does fail for someone, loop a few times after load checking.
	if (not Multimod.homeServer or Multimod.homeServer == nil or Multimod.homeServer == '') then
		C_Timer.After(5, function()
			Multimod:setHomeServer()
		end)
	end
end

--A short delay after logon for commands that require item cache loaded properly.
Multimod.startDelaySeconds = 30
local f = CreateFrame("Frame")
f:RegisterEvent("PLAYER_LOGIN")
f:SetScript("OnEvent", function(self, event, ...)
	--Create a timestamp delay for commands that require item cache loaded properly.
	ShowFriends()
	Multimod.startDelay = time() + Multimod.startDelaySeconds
	Multimod:chatPrefixMsg("Multimod v" .. GetAddOnMetadata("Multimod", "Version") .." loaded.")
	--Make sure equipped items are cached.
	C_Timer.After(Multimod.startDelaySeconds, function()
		Multimod:buildItemCache()
	end)
	--Run perform at logon options 10 secs after we start loading item cache just to be sure it's finished.
	C_Timer.After(Multimod.startDelaySeconds + 10, function()
		Multimod:logon_functions()
	end)
	--Backup join addon channel timer.
	C_Timer.After(45, function()
		Multimod:join_addoncomms()
	end)
	Multimod:createClickButtons()
end)

--Things to run shortly after logon as specified in options (after item cache loads).
function Multimod:logon_functions()
	--Run a backup itembuttons update, sometimes it fails at login, probably becaus of itemcache load times.
	Multimod:updateItemButtons()
	if (Multimod.db.global.logon_token) then
		SlashCmdList.MMTOKENCMD()
	end
	if (Multimod.db.global.logon_invasion) then
		Multimod:invasionTimer()
	end
	if (Multimod.db.global.logon_delete_argus) then
		SlashCmdList.DESTROYARGUSCMD("lowSpam")
	end
	if (Multimod.db.global.logon_delete_armortokens) then
		SlashCmdList.DESTROYATCMD("lowSpam")
	end
	if (Multimod.db.global.logon_delete_love) then
		SlashCmdList.DESTROYLOVECMD("lowSpam")
	end
	if (Multimod.db.global.logon_delete_halloween) then
		SlashCmdList.DESTROYHALLOWEENCMD("lowSpam")
	end
	if (Multimod.db.global.logon_delete_xmas) then
		SlashCmdList.DESTROYXMASCMD("lowSpam")
	end
	if (Multimod.db.global.logon_scan) then
		scanInventory()
	end
	if (Multimod.db.global.logon_passloot) then
		SlashCmdList.PASSLOOTONCMD()
	end
end

----------------------
--Addon Comm Channel--
----------------------

--Join addon channel a few seconds after joining default channels at logon.
--There is also a delayed join timer in PLAYER_LOGIN event just incase they don't join any channels.
local f = CreateFrame("Frame")
f:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")
f:RegisterEvent("CHANNEL_UI_UPDATE")
f:SetScript('OnEvent', function(self, event, ...)
	if (GetNumDisplayChannels() > 0) then
		self:UnregisterEvent("CHAT_MSG_CHANNEL_NOTICE")
		self:UnregisterEvent("CHANNEL_UI_UPDATE")
		ShowFriends()
		C_Timer.After(8, function()
			Multimod:join_addoncomms()
		end)
	end
end)

function Multimod:join_addoncomms(force)
	--Only join if atleast the default channels have loaded
	local _, _, _, secondChannel = GetChannelList()
	if (secondChannel ~= nil or force  ~= nil) then
		JoinTemporaryChannel(Multimod.addonChannel)
		createMapCache()
	else
		--Force join in 2 minutes anyway just incase it's a rare user than removed the main chat channels General etc.
		C_Timer.After(120, function()
			Multimod:join_addoncomms(true)
		end)
	end
end

function Multimod:OnCommReceived(commPrefix, string, distribution, sender)
	if (commPrefix ~= Multimod.commPrefix) then
		return
	end
	--Valid distribution types: "PARTY", "RAID", "GUILD", "OFFICER", "BATTLEGROUND", "WHISPER", "CHANNEL"
	local args = explode(" ", string, 1)
	local dataPrefix = args[1] --Prefix so we know where to send the data
	local dataString = args[2] --All characters after the prefix
	if (dataPrefix == "ksupdate") then
		--Keystone update
		local classAndKey = explode(",", dataString, 2)
		local class = classAndKey[1]
		local kstime = classAndKey[2]
		local keystone = classAndKey[3]
		receiveKeystoneParty(sender, class, kstime, keystone)
	end
	if (dataPrefix == "mmmap") then
    	--Map update
    	mapCommReceived(sender, dataString, distribution)
    end
end

--Ace3 doesn't support sending bnet msgs yet so this is done without using the lib.
local f = CreateFrame("Frame")
f:RegisterEvent("BN_CHAT_MSG_ADDON")
f:SetScript('OnEvent', function(self, event, addonPrefix, string, distribution, senderPresenceID)
	--distribution always == "WHISPER" for bnet event, but I change it here to tell the diff.
	local args = explode(" ", string, 1)
	local dataPrefix = args[1] --Prefix so we know where to send the data
	local dataString = args[2] --All characters after the prefix
	if (dataPrefix == "ksupdate") then
		--Keystone update
		local classAndKey = explode(",", dataString, 2)
		local class = classAndKey[1]
		local kstime = classAndKey[2]
		local keystone = classAndKey[3]
		receiveKeystoneParty(sender, class, kstime, keystone)
	end
	if (dataPrefix == "mmmap") then
    	--Map update
    	mapCommReceived(senderPresenceID, dataString, "BNET")
    end
end)

function Multimod:sendAddonChannel(distribution, string, target)
	if distribution == "CHANNEL" then
		--Get channel ID number.
		local addonChannelId = GetChannelName(Multimod.addonChannel)
		--Not sure why this only accepts a string and not an int
		target = tostring(addonChannelId)
	elseif distribution ~= "WHISPER" then
		target = nil
	end
	Multimod:SendCommMessage(Multimod.commPrefix, string, distribution, target)
end

--------
--Misc--
--------

--Printing msgs to the default chat window.
Multimod.chatPrefix = "ff8700"
function Multimod:loadChatColor()
	Multimod.chatColor = "|cff" .. RGBToHex(Multimod.db.global.chat_msg_color_r, Multimod.db.global.chat_msg_color_g, Multimod.db.global.chat_msg_color_b)
end
function Multimod:chatMsg(msg, addPrefix)
	--print(gsub(chatColor, "\124", "\124\124"))
	if (addPrefix) then
		print("|cff" .. Multimod.chatPrefix .. "(MM):|r " .. Multimod.chatColor .. msg)
	else
		print(Multimod.chatColor .. msg)
	end
end

--Same as above Multimod:chatMsg() but adds prefix.
function Multimod:chatPrefixMsg(msg, color)
	Multimod:chatMsg(msg, true)
end

--Run all destroy funcs in one command, xmas junk, halloween junk, argus junk etc
SLASH_DESTROYALLCMD1, SLASH_DESTROYALLCMD2 = '/destroyall', '/deleteall';
function SlashCmdList.DESTROYALLCMD(msg, editBox)
	if (Multimod.startDelay > time()) then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	Multimod:chatPrefixMsg("|cff" .. Multimod.chatPrefix .. "Running all destroy commands.")
	SlashCmdList.DESTROYLOVECMD("lowSpam") --Love is in the air items
	SlashCmdList.DESTROYHALLOWEENCMD("lowSpam") --Halloween items
	SlashCmdList.DESTROYXMASCMD("lowSpam") --Winter Veil items
	SlashCmdList.DESTROYARGUSCMD("lowSpam") --Argus items
	SlashCmdList.DESTROYATCMD("lowSpam") --Unsullied and Dauntless armor tokens
	Multimod:chatPrefixMsg("|cff" .. Multimod.chatPrefix .. "Finished running all destroy commands.")
end

--Auto insert keystones
local f = CreateFrame("Frame")
f:RegisterEvent("CHALLENGE_MODE_KEYSTONE_RECEPTABLE_OPEN")
f:SetScript("OnEvent", function(self, event, ...)
	if (Multimod.db.global.insert_keystones) then
		C_Timer.After(0.5, function()
			insertKeystone()
		end)
	end
end)

--Auto add the DELETE text to delete item confirmation popups.
local f = CreateFrame("Frame")
f:RegisterEvent("DELETE_ITEM_CONFIRM")
f:SetScript("OnEvent", function(self, event, ...)
	if (Multimod.db.global.auto_delete_text) then
		for i = 1, STATICPOPUP_NUMDIALOGS do
			local dialogType = _G["StaticPopup"..i].which
			if (dialogType == "DELETE_GOOD_ITEM") then
				_G["StaticPopup" .. i .. "EditBox"]:SetText(DELETE_ITEM_CONFIRM_STRING)
			end
		end
	end
end)

--Auto close the popup window you get after joining a raid group that shows your role (custom group finder etc).
local f = CreateFrame("Frame")
f:RegisterEvent("LFG_LIST_JOINED_GROUP")
f:SetScript("OnEvent", function(self, event, ...)
	--If group leader or not in group the popup looks different (same frame name though).
	if (UnitIsGroupLeader("player") or not IsInGroup()) then
		return
	end
	if (Multimod.db.global.role_confirm) then
		local roleText = LFGListInviteDialog.Role:GetText()
		if (roleText and roleText ~= "") then
			Multimod:chatPrefixMsg("Joined group as role:|cffFFFF00 " .. roleText)
		end
		LFGListInviteDialog:Hide()
	end
end)

--Auto accept role check
local roleCheckMsg = true
local f = CreateFrame("Frame")
f:RegisterEvent("LFG_ROLE_CHECK_SHOW")
f:SetScript("OnEvent", function(self, event, ...)
	local queueText = LFDRoleCheckPopupDescriptionText:GetText()
	local doQueue = false
	if (Multimod.db.global.role_check) then --Global setting
		doQueue = true
	elseif (string.find(queueText, "The Crown Chemical Co") and Multimod.db.global.role_check_love) then --Love is in the air
		doQueue = true
	elseif (string.find(queueText, "The Headless Horseman") and Multimod.db.global.role_check_halloween) then --Halloween
		doQueue = true	--(untested, hopefully this is the dung name)
	elseif (string.find(queueText, "Timewalking") and Multimod.db.global.role_check_timewalking) then --Timewalking
		doQueue = true	--(untested, hopefully timewalking is in the queue name)
	end
	leader, tank, healer, damage = GetLFGRoles()
	if (tank == true) then
		tankMsg = "Tank "
	else
		tankMsg = ""
	end
	if (healer == true) then
		healerMsg = "Healer "
	else
		healerMsg = ""
	end
	if (damage == true) then
		damageMsg = "Damage"
	else
		damageMsg = ""
	end
	if (doQueue and roleCheckMsg == true) then
		CompleteLFGRoleCheck(true)
		Multimod:chatPrefixMsg("Auto completed role check as:|cffFFFF00 " .. tankMsg .. healerMsg .. damageMsg)
		--Turn off the msg for a few seconds, LFG_ROLE_CHECK_SHOW seems to trigger when anyone in group signs up so it spams.
		--There's probably a better event to make it only trigger on myself somehow, will work it out later.
		roleCheckMsg = false
		C_Timer.After(5, function()
			resetRoleCheckMsg()
		end)
	end
end)

function resetRoleCheckMsg()
	roleCheckMsg = true
end

--Leave group, if in a dungeon/raid finder group it will leave that group first but keep real group intact.
--This is named "/done" becaus I would use it when each dungeon was done when leveling.
SLASH_DONECMD1 = '/done';
function SlashCmdList.DONECMD(msg, editBox)
	LeaveParty();
end

--These /out and /in commands and good for handing in quests back at the start of dungeons while leveling.
--Teleport out of dungeon (same as clicking the minimap icon and choosing teleport out of dungeon)
SLASH_OUTCMD1 = '/out';
function SlashCmdList.OUTCMD(msg, editBox)
	LFGTeleport(true);
end

--Teleport to dungeon (same as clicking the minimap icon and choosing teleport to dungeon)
SLASH_INCMD1 = '/in';
function SlashCmdList.INCMD(msg, editBox)
	LFGTeleport();
end

--Turn on pvp flag
SLASH_PVPONCMD1 = '/pvpon';
function SlashCmdList.PVPONCMD(msg, editBox)
	SetPVP(1);
end

--Turn off pvp flag
SLASH_PVPOFFCMD1 = '/pvpoff';
function SlashCmdList.PVPOFFCMD(msg, editBox)
	SetPVP();
end

--To reset cam potion, most people have a macro for this but I threw it in anyway.
SLASH_CAMCMD1 = '/cam';
function SlashCmdList.CAMCMD(msg, editBox)
	SetView(5);SetView(5)
end

--Closes all windows, same as Escape but doesn't open the game menu, handy when chars are out of sync.
SLASH_CLOSEALLCMD1 = '/closeall';
function SlashCmdList.CLOSEALLCMD(msg, editBox)
	CloseAllWindows()
end

--Sells all greys, but only if shop is open so you don't accidently equip them.
SLASH_GREYSCMD1 = '/greys';
function SlashCmdList.GREYSCMD(msg, editBox)
	if MerchantFrame:IsVisible() then
		local c,i,n,v=0;for b=0,4 do 
			for s=1,GetContainerNumSlots(b)do
				i={GetContainerItemInfo(b,s)}n=i[7]if n and string.find(n,"9d9d9d")then 
					v={GetItemInfo(n)}q=i[2]c=c+v[11]*q;UseContainerItem(b,s)print(n,q)
				end
			end
		end
		print(GetCoinText(c))
	else
		Multimod:chatPrefixMsg("Selling greys halted, shop not open.")
	end
end

--Pass on loot rolls, much easier to be passing on loot on slave chars while in leveling dungeons.
SLASH_PASSLOOTONCMD1, SLASH_PASSLOOTONCMD2 = '/passlooton', '/passloot';
function SlashCmdList.PASSLOOTONCMD(msg, editBox)
	SetOptOutOfLoot(true)
end

--Stop passing on loot
SLASH_PASSLOOTOFFCMD1, SLASH_PASSLOOTOFFCMD2 = '/passlootoff', '/passlootstop';
function SlashCmdList.PASSLOOTOFFCMD(msg, editBox)
	SetOptOutOfLoot()
end

SLASH_BNETINVITECMD1, SLASH_BNETINVITECMD2 = '/bnetinvite', '/bnetinv';
function SlashCmdList.BNETINVITECMD(msg, editBox)
	ShowFriends()
	local presenceID = GetAutoCompletePresenceID(msg)
	if (not msg or msg == '') then
		Multimod:chatPrefixMsg("Please specify a bnet friend name to invite all online characters from.")
	elseif (not presenceID or presenceID == '') then
		Multimod:chatPrefixMsg("Battle.net friend doesn't exist: " .. msg .. ".")
	else
		local friendIndex = BNGetFriendIndex(presenceID)
		local i = 1
		local _, toonName = BNGetFriendGameAccountInfo(friendIndex, i)
		if (not toonName or toonName == '') then
			Multimod:chatPrefixMsg("Battle.net friend " .. msg .. " has no characters online.")
		else
    		while toonName do
				if (toonName and toonName ~= '') then
					InviteUnit(toonName)
				end
      			i = i + 1
      			_, toonName = BNGetFriendGameAccountInfo(friendIndex, i)
    		end
    		Multimod:chatPrefixMsg("Invited " .. i-1 .. " characters of " .. msg .. ".")
		end
	end
end

--This spams follow on the name you specify every 2 seconds, not really needed if you use Jamba.
--I put this in anyway for some friends that would follow me through dungeons semi afk a while back.
--Usage: "/followspam name" to start and "/followspam stop" to stop.
local followspam = CreateFrame("Frame");
local followspam_delay=2
SLASH_FOLLOWSPAMCMD1, SLASH_FOLLOWSPAMCMD2 = '/followspam', '/spamfollow';
function SlashCmdList.FOLLOWSPAMCMD(msg, editBox)
	msg = string.lower(msg)
	if msg == "stop" then
		followspam:SetScript("OnUpdate", nil);
	else
		local follow_who = msg
		followspam:SetScript("OnUpdate", function(self, expired)
			followspam_delay = followspam_delay - expired
			if followspam_delay < 0 then
				followspam_delay = 2
				FollowUnit(follow_who, true);
			end
		end);	
    end
end

--Destroy argus eggs, they are just an annoyance that you can't vendor once you get all the mounts.
SLASH_DESTROYEGGSCMD1, SLASH_DESTROYEGGSCMD2 = '/destroyeggs', '/deleteggs';
function SlashCmdList.DESTROYEGGSCMD(msg, editBox)
	if (Multimod.startDelay > time() and msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	Multimod:chatPrefixMsg("Destroying all argus eggs...")
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local name=GetContainerItemLink(bag,slot)
			if (name and string.find(name,"Fel%-Spotted Egg")) then
				PickupContainerItem(bag,slot);DeleteCursorItem()
				Multimod:chatMsg("Destroyed: " .. name .. "|cff00FF00.")
			end
		end
	end
	Multimod:chatMsg("Done.")
end

--Destroy argus eggs as well as all the duplicate mounts/pets/toys.
SLASH_DESTROYARGUSCMD1, SLASH_DESTROYARGUSCMD2 = '/destroyargus', '/deleteargus';
function SlashCmdList.DESTROYARGUSCMD(msg, editBox)
	if (Multimod.startDelay > time() and msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	if (msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Destroying all non-sellable |cffFFFF00Argus mounts/pets/toys|cff00FF00...")
	end
	local a = {}
	--Eggs
	if (Multimod.db.global.argus_eggs) then
		table.insert(a, "Fel%-Spotted Egg")
	end
	--Muffins
	if (Multimod.db.global.argus_muffins) then
		table.insert(a, "Carefully Hidden Muffin")
	end
	--Random junk
	if (Multimod.db.global.argus_radinax_gem) then
		table.insert(a, "Cracked Radinax Control Gem")
	end
	--Cycle mount list and insert into the array any argus mounts you already own.
	for i = 1, GetNumCompanions("MOUNT") do
    	local name = C_MountJournal.GetDisplayedMountInfo(i)
    	--Egg mount drops
    	if (Multimod.db.global.argus_darkspore and name == "Darkspore Mana Ray") then
    		table.insert(a, "Darkspore Mana Ray")
		elseif (Multimod.db.global.argus_felglow and name == "Felglow Mana Ray") then
			table.insert(a, "Felglow Mana Ray")
		elseif (Multimod.db.global.argus_scint and name == "Scintillating Mana Ray") then
			table.insert(a, "Scintillating Mana Ray")
		elseif (Multimod.db.global.argus_vibrant and name == "Vibrant Mana Ray") then
			table.insert(a, "Vibrant Mana Ray")
		--Mac'Aree drops
		elseif (Multimod.db.global.argus_maddened and name == "Maddened Chaosrunner") then
			table.insert(a, "Maddened Chaosrunner")
		elseif (Multimod.db.global.argus_lambent and name == "Lambent Mana Ray") then
			table.insert(a, "Lambent Mana Ray")
		elseif (Multimod.db.global.argus_acid and name == "Acid Belcher") then
			table.insert(a, "Acid Belcher")
		--Antoran Wastes drops
		elseif (Multimod.db.global.argus_crimson and name == "Crimson Slavermaw") then
			table.insert(a, "Crimson Slavermaw")
		elseif (Multimod.db.global.argus_vile and name == "Vile Fiend") then
			table.insert(a, "Vile Fiend")
		end
	end
	--Check if toys are owned, insert into array if we already have it.
    if (Multimod.db.global.argus_baarut and PlayerHasToy(153193)) then
    	table.insert(a, "Baarut the Brisk")
	end
	if (Multimod.db.global.argus_barrier and PlayerHasToy(153183)) then
    	table.insert(a, "Barrier Generator")
	end	
	if (Multimod.db.global.argus_orb and PlayerHasToy(153194)) then
    	table.insert(a, "Legion Communication Orb")
	end	
	if (Multimod.db.global.argus_blue and PlayerHasToy(153179)) then
    	table.insert(a, "Blue Conservatory Scroll")
	end	
	if (Multimod.db.global.argus_red and PlayerHasToy(153181)) then
    	table.insert(a, "Red Conservatory Scroll")
	end	
	if (Multimod.db.global.argus_yellow and PlayerHasToy(153180)) then
    	table.insert(a, "Yellow Conservatory Scroll")
	end	
	if (Multimod.db.global.argus_sfe and PlayerHasToy(153253)) then
    	table.insert(a, "S.F.E. Interceptor")
	end	
	if (Multimod.db.global.argus_eye and PlayerHasToy(153293)) then
    	table.insert(a, "Sightless Eye")
	end	
	if (Multimod.db.global.argus_spire and PlayerHasToy(153124)) then
    	table.insert(a, "Spire of Spite")
	end	
	if (Multimod.db.global.argus_micro and PlayerHasToy(153126)) then
    	table.insert(a, "Micro-Artillery Controller")
	end	
	if (Multimod.db.global.argus_holy and PlayerHasToy(153182)) then
    	table.insert(a, "Holy Lightsphere")
	end
	--Check if pets are owned, insert into array if we already have it (no point getting 3 of these they are worthess).
	local numPetsInGame, numPetsOwned = C_PetJournal.GetNumPets()
	local insert_docile = false --We have these vars so it's only inserted once if more than one of the same pet is collected.
	local insert_fel = false
	local insert_grasping = false
	local insert_fel_lasher = false
	local insert_uunas_doll = false
	for i = 1, numPetsOwned do
    	local _, _ ,_, _, _, _, _, name = C_PetJournal.GetPetInfoByIndex(i)
		if (Multimod.db.global.argus_docile and name == "Docile Skyfin" and not insert_docile) then
			table.insert(a, "Docile Skyfin")
			insert_docile = true
		end
		if (Multimod.db.global.argus_felafflicted and name == "Fel-Afflicted Skyfin"  and not insert_fel) then
			table.insert(a, "Fel%-Afflicted Skyfin")
			insert_fel = true
		end
		if (Multimod.db.global.argus_grasping and name == "Grasping Manifestation" and not insert_grasping) then
			table.insert(a, "Grasping Manifestation")
			insert_grasping = true
		end
		if (Multimod.db.global.argus_fel_lasher and name == "Fel_lasher" and not insert_fel_lasher) then
			table.insert(a, "Fel_lasher")
			insert_fel_lasher = true
		end
		if (Multimod.db.global.argus_uunas_doll and name == "Uuna" and not insert_uunas_doll) then
			table.insert(a, "Uuna%'s Doll")
			insert_uunas_doll = true
		end
	end
	if (next(a) == nil and msg ~= "lowSpam") then
   		Multimod:chatMsg("All Argur items are unticked in options, nothing to destroy.")
   		return
	end
	local count = 0
	--Destroy all items in the array.
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
    		for i, v in ipairs(a) do
      			local name=GetContainerItemLink(bag,slot)
				if (name and string.find(name, v) and name ~= nill and name ~= '') then
					PickupContainerItem(bag,slot)
					DeleteCursorItem()
					Multimod:chatMsg("Destroyed: " .. name .. "|cff00FF00.")
					count = count + 1
				end
			end
		end
	end
	if (msg ~= "lowSpam") then
		Multimod:chatMsg("Done, " .. count .. " items found and destroyed.")
	end
end

--Delete all argus armor tokens ticked in options, if shop is open we sell them instead (only worth like 2-3g each).
SLASH_DESTROYATCMD1, SLASH_DESTROYATCMD2 = '/destroyat', '/deleteat';
function SlashCmdList.DESTROYATCMD(msg, editBox)
	if (Multimod.startDelay > time() and msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	if (msg ~= "lowSpam") then
		if MerchantFrame:IsVisible() then
			Multimod:chatPrefixMsg("Selling all argus armor tokens...")
		else
			Multimod:chatPrefixMsg("Destroying all argus armor tokens...")
		end
	end
	local a = {}
	--Unsullied rings
	if (Multimod.db.global.delete_unsullied_ring) then
		table.insert(a, "Unsullied Ring")
	end
	--Unsullied necks
	if (Multimod.db.global.delete_unsullied_necklace) then
		table.insert(a, "Unsullied Necklace")
	end
	--Unsullied trinkets
	if (Multimod.db.global.delete_unsullied_trinket) then
		table.insert(a, "Unsullied Trinket")
	end
	--Unsullied cloaks
	if (Multimod.db.global.delete_unsullied_cloak) then
		table.insert(a, "Unsullied Cloak")
	end
	--Unsullied relics
	if (Multimod.db.global.delete_unsullied_relic) then
		table.insert(a, "Unsullied Relic")
	end
	--Unsullied cloth
	if (Multimod.db.global.delete_unsullied_cloth) then
		table.insert(a, "Unsullied Cloth")
	end
	--Unsullied leather
	if (Multimod.db.global.delete_unsullied_leather) then
		table.insert(a, "Unsullied Leather")
	end
	--Unsullied mail
	if (Multimod.db.global.delete_unsullied_mail) then
		table.insert(a, "Unsullied Mail")
	end
	--Unsullied plate
	if (Multimod.db.global.delete_unsullied_plate) then
		table.insert(a, "Unsullied Plate")
	end
	--850 tokens from the broken shore
	if (Multimod.db.global.delete_dauntless_tokens) then
		table.insert(a, "Dauntless Bracers")
		table.insert(a, "Dauntless Choker")
		table.insert(a, "Dauntless Cloak")
		table.insert(a, "Dauntless Gauntlets")
		table.insert(a, "Dauntless Girdle")
		table.insert(a, "Dauntless Hood")
		table.insert(a, "Dauntless Leggings")
		table.insert(a, "Dauntless Ring")
		table.insert(a, "Dauntless Spaulders")
		table.insert(a, "Dauntless Treads")
		table.insert(a, "Dauntless Trinket")
		table.insert(a, "Dauntless Tunic")
	end
	if (next(a) == nil and msg ~= "lowSpam") then
   		Multimod:chatMsg("All Argus Armor Tokens items are unticked in options, nothing to destroy.")
   		return
	end
	local count = 0
	--Destroy all items in the array.
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
    		for i, v in ipairs(a) do
      			local name=GetContainerItemLink(bag,slot)
				if (name and string.find(name, v) and name ~= nill and name ~= '') then
					if MerchantFrame:IsVisible() then
						--850 Dauntless can't be sold so we destroy it even with shop open.
						if (string.find(name, "Dauntless")) then
							PickupContainerItem(bag,slot)
							DeleteCursorItem()
							Multimod:chatMsg("Destroyed: " .. name .. "|cff00FF00.")
						else
							UseContainerItem(bag,slot)
							Multimod:chatMsg("Sold: " .. name .. "|cff00FF00.")
						end
					else
						PickupContainerItem(bag,slot)
						DeleteCursorItem()
						Multimod:chatMsg("Destroyed: " .. name .. "|cff00FF00.")
					end
					count = count + 1
				end
			end
		end
	end
	if (msg ~= "lowSpam") then
		Multimod:chatMsg("Done, " .. count .. " items found and destroyed.")
	end
end

--Turn Lua errors on and off
SLASH_LUAONCMD1 = '/luaon';
function SlashCmdList.LUAONCMD(msg, editBox)
	SetCVar("ScriptErrors","1")
	Multimod:chatPrefixMsg("Lua errors enabled.")
end

SLASH_LUAOFFCMD1 = '/luaoff';
function SlashCmdList.LUAOFFCMD(msg, editBox)
	SetCVar("ScriptErrors","0")
	Multimod:chatPrefixMsg("Lua errors disabled.")
end

--Invasion timer (8.0 prepatch rotation).
SLASH_INVASIONTIMERCMD1 = '/invasion';
function SlashCmdList.INVASIONTIMERCMD(msg, editBox)
	Multimod:invasionTimer()
end

--Invasion timer (guild msg).
SLASH_GINVASIONTIMERCMD1 = '/ginvasion';
function SlashCmdList.GINVASIONTIMERCMD(msg, editBox)
	Multimod:invasionTimer("guild")
end

function Multimod:invasionTimer(msgType)
	--Invasion start time from the past to calculate from, cycle is 6 hours on, 12.5 hours off
	--Old pre 8.0 start time. local time_start = 1506389400 --09/26/2017 @ 1:30am (UTC)
	local time_start = 1532430000 --07/24/2018 @ 11:00am (UTC)
	local time_finish = time_start + 21600
	local time = time()
	local i = 0
	while (i < 9999) do
		i = i + 1;
		--Get the closest future timestamp of invasion start
		if (time < time_start) then
			time_left = getTimeString(time_start);
			if (msgType == "guild") then
				SendChatMessage("(MM): The next Legion Invasion will start in " .. time_left .. ".", "guild")
			else
				Multimod:chatPrefixMsg("The next Legion Invasion will start in " .. time_left .. ".")
			end
			return
		end
		if (time < time_finish) then
			time_left = getTimeString(time_finish);
			if (msgType == "guild") then
				SendChatMessage("(MM): The current Legion Invasion will end in " .. time_left .. ".", "guild")
			else
				Multimod:chatPrefixMsg("The current Legion Invasion will end in " .. time_left .. ".")
			end
			return
		end
		--Add 18.5h to the time for a full cycle
		time_start = time_start + 66600
		time_finish = time_finish + 66600
	end
end

--Display current token price
local mmShowTokenPrice = false
local mmShowTokenGuild = false
SLASH_MMTOKENCMD1, SLASH_MMTOKENCMD2 = '/token', '/tokens';
function SlashCmdList.MMTOKENCMD(msg, editBox)
	C_WowTokenPublic.UpdateMarketPrice();
	--Using this variable so it doesn't msg chat when you open AH or other token price updates.
	mmShowTokenPrice = true
end

--Show price to guild.
SLASH_MMGTOKENCMD1, SLASH_MMGTOKENCMD2 = '/gtoken', '/gtokens';
function SlashCmdList.MMGTOKENCMD(msg, editBox)
	C_WowTokenPublic.UpdateMarketPrice();
	mmShowTokenGuild = true
end

local mmTokenPrice = CreateFrame("Frame")
mmTokenPrice:RegisterEvent("TOKEN_MARKET_PRICE_UPDATED")
mmTokenPrice:SetScript("OnEvent", function(self, event, ...)
	if (mmShowTokenPrice) then
		local currentPrice = C_WowTokenPublic.GetCurrentMarketPrice()
		if (not currentPrice or currentPrice == nil or currentPrice == '' or currentPrice == 0) then
			Multimod:chatPrefixMsg("The Blizzard token API is down.")
			return
		end
		Multimod:chatPrefixMsg("Current token price:|cffFFFF00 " .. comma_value(GetMoneyString(currentPrice)) .. "")
		mmShowTokenPrice = false
	end
	if (mmShowTokenGuild) then
		local currentPrice = C_WowTokenPublic.GetCurrentMarketPrice()
		if (not currentPrice or currentPrice == nil or currentPrice == '' or currentPrice == 0) then
			Multimod:chatPrefixMsg("The Blizzard token API is down.")
			return
		end
		currentPrice = string.sub(currentPrice, 1, -5)
		SendChatMessage("(MM): Current token price: " .. comma_value(currentPrice), "guild")
		mmShowTokenGuild = false
	end
end)

--Grant a level, there's an option in /mm to auto accept granted levels.
SLASH_GRANTLEVELCMD1 = '/grant';
function SlashCmdList.GRANTLEVELCMD(msg, editBox)
	local targetName = UnitName("target")
	if (targetName ~= nil and targetName ~= '') then
		Multimod:chatPrefixMsg("Granting level to:|cffFFFF00 " .. targetName)
		GrantLevel("target", 1)
	else
		Multimod:chatPrefixMsg("You need a target to grant a level.")
	end
end

local f = CreateFrame("Frame")
f:RegisterEvent("LEVEL_GRANT_PROPOSED")
f:SetScript("OnEvent", function(self, event, ...)
	if (Multimod.db.global.level_granted) then
		AcceptLevelGrant()
	end
end)

----------------------------------
--Scan your bags for higher gear--
----------------------------------

--Had to add this cache rather than query from the client, they renamed stuff in cata and there are name mismatches between diff funcs.
local equipmentSlotCache = {
	["INVTYPE_HEAD"] = 1,
	["INVTYPE_NECK"] = 2,
	["INVTYPE_SHOULDER"] = 3,
	["INVTYPE_BODY"] = 4,
	["INVTYPE_CHEST"] = 5,
	["INVTYPE_ROBE"] = 5,
	["INVTYPE_WAIST"] = 6,
	["INVTYPE_LEGS"] = 7,
	["INVTYPE_FEET"] = 8,
	["INVTYPE_WRIST"] = 9,
	["INVTYPE_HAND"] = 10,
	["INVTYPE_FINGER"] = 11,
	["INVTYPE_TRINKET"] = 13,
	["INVTYPE_CLOAK"] = 15,
	["INVTYPE_WEAPON"] = 16,
	["INVTYPE_2HWEAPON"] = 16,
	["INVTYPE_WEAPONMAINHAND"] = 16,
	["INVTYPE_HOLDABLE"] = 16,
	["INVTYPE_RANGED"] = 16,
	["INVTYPE_SHIELD"] = 17,
	["INVTYPE_WEAPONOFFHAND"] = 17,
	["INVTYPE_RANGEDRIGHT"] = 17,
}

--Scan and print to chat any higher ilvl items in inventory than you have equipped.
--It reports legendaries as being higher even though you can only equip 2, but this won't be an isssue when BFA launches.
SLASH_SCANHIGHCMD1 = '/scan';
function SlashCmdList.SCANHIGHCMD(msg, editBox)
	scanInventory(msg)
end
--Same scan as above but also equips the higher item.
SLASH_EQUIPHIGHERCMD1 = '/scanequip';
function SlashCmdList.EQUIPHIGHERCMD(msg, editBox)
	scanInventory("equip")
end

--This founc isn't the prettiest way of doing this but looping the inventory once is more efficient than looping the
--slots once then checking the inventory for every slot. It's already instant so I'm not going to bother building an inventory cache.
--Rings and trinkets are done in thier own function below this one.
--TODO: Add atribute checking so we equip the right sort over everything else the same ilvl, agi for agi classes etc.
--TODO: Some leveling dung drops give the wrong attributes especially the dungeon bags.
function scanInventory(msg)
	if (Multimod.startDelay > time() and msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	if (msg == "equip") then
		Multimod:chatPrefixMsg("Scanning inventory for higher ilvl items to |cffff8700equip|cff00FF00...")
		if MerchantFrame:IsVisible() then
			Multimod:chatPrefixMsg("Halted scanning bags for items to equip while shop is open to prevent you accidently selling things.")
			return
		end
	else
		Multimod:chatPrefixMsg("Scanning inventory for higher ilvl equipment...")
	end
	mmScanCount = 0
	local skip = {}
	for bag = 0, NUM_BAG_SLOTS do
		bagTemp = bag
		for slot = 1, GetContainerNumSlots(bag) do
			slotTemp = slot
			--Unset variables at start of the loop
			--Shouldn't really need to unset these but if an item query fails it's better the loop skips it instead of showing wrong item.
			local itemName = nil
			local itemLink = nil
			local itemLevel = nil
			local itemMinLevel = nil
			local itemEquipLoc = nil
			local itemLevel = nil
			local equippedItemName = nil
			local equippedItemLink = nil
			local equippedItemLevel = nil
			local equippedItemMinLevel = nil
			local equippedItemEquipLoc = nil
			local equippedItemLevel = nil
			local slotNumber = nil
			local _, _, _, _, _, _, itemLink, _, _, itemId = GetContainerItemInfo(bag, slot)
			if (itemId ~= nil and itemId ~= '') then
				local itemName, itemLink, itemQuality, itemLevel, itemMinLevel, _, _, _, itemEquipLoc = GetItemInfo(itemLink)
				--Get real item level from the visual link GetItemInfo() returns wrong ilvl on timewalking gear and some other stuff.
				itemLevel = getRealItemLvl(itemLink)
				if (itemEquipLoc ~= nil and itemEquipLoc ~= '') then
					--If item is equippable then get real item level and check the corresponding equipped slot
					local slotNumber = equipmentSlotCache[itemEquipLoc] --Local cache more reliable than client query, things mismatch
					local equippedItemId = GetInventoryItemID("player", slotNumber)
					--Get item level and info for equipped item in slot matching where current container item would equip to.
					local equippedItemLink = GetInventoryItemLink("player", slotNumber)
					if (equippedItemId ~= nil and equippedItemId ~= '') then
						equippedItemName, equippedItemLink, equippedItemQuality, equippedItemLevel,
								equippedItemMinLevel, _, _, _, equippedItemEquipLoc = GetItemInfo(equippedItemLink)
						equippedItemLevel = getRealItemLvl(equippedItemLink)
					end
					if (itemLevel ~= nil and itemLevel ~= '' and slotNumber ~= nil and slotNumber ~= '') then
						itemLocName = _G[itemEquipLoc]
						if (not equippedItemLevel or equippedItemLevel == nil or equippedItemLevel == '') then
							equippedItemLevel = 0
							equippedMsg = "(empty slot)"
						else
							equippedMsg = equippedItemLevel
						end
						if (slotNumber == 11 and not skip[slotNumber]) then
							--If a ring is found, go to the ring func to make sure we equip the highest in both slots.
							scanBothSlots(msg, slotNumber, equippedItemLevel, equippedItemLink)
							skip[slotNumber] = true
						end
						if (slotNumber == 13 and not skip[slotNumber]) then
							--If a trinket is found, go to the trinket func to make sure we equip the highest in both slots.
							scanBothSlots(msg, slotNumber, equippedItemLevel, equippedItemLink)
							skip[slotNumber] = true
						end
						--Ignoring artifacts(6) and heirlooms(7), will remove artifact check in BFA when we have no offspec artifacts.
						if (itemLevel > equippedItemLevel and equippedItemQuality ~= 6 and equippedItemQuality ~= 7
								and not skip[slotNumber]) then
							found = true
							--foundCount = foundCount + 1
							if (not equippedItemLevel or equippedItemLevel == nil or equippedItemLevel == '' or equippedItemLevel == 0) then
								equippedItemLevel = 0
								equippedMsg = "(empty slot)"
							else
								equippedMsg = equippedItemLevel
							end
							--If there are multiple items for this slot in bags then get the highest.
							highestItemLink, highestItemLevel, highestBag, highestSlot= getHighestItemPerSlot(slotNumber)
							--If highest is found then update the variables and add this slot to be skipped for future loops.
							--If for some reason highest func fails it will equip the current item and next highest found next loop.
							--That's why the foundScannedUpgrade() is below this "if" and not in it.
							if (itemLevel >= highestItemLevel) then
								itemLevel = highestItemLevel
								itemLink = highestItemLink
								bagTemp = highestBag
								slotTemp = highestSlot
								skip[slotNumber] = true
							end
							foundScannedUpgrade(msg, bag, slot, itemLevel, itemLink, itemEquipLoc, equippedMsg)
						end
					end
				end
			end
		end
	end
	if (found) then
		if (foundCount == 1) then
			Multimod:chatMsg("Done, " .. mmScanCount .. " item found.")
		else
			Multimod:chatMsg("Done, " .. mmScanCount .. " items found.")
		end
	else
		Multimod:chatMsg("No higher items found.")
	end
end

function scanBothSlots(msg, slotNumber, equippedItemLevel, equippedItemLink)
	if (slotNumber == 11) then
		slotOneItemLink = GetInventoryItemLink("player", 11)
		slotTwoItemLink = GetInventoryItemLink("player", 12)
	else
		slotOneItemLink = GetInventoryItemLink("player", 13)
		slotTwoItemLink = GetInventoryItemLink("player", 14)
	end
	local slotOneItemLevel = getRealItemLvl(slotOneItemLink)
	local slotTwoItemLevel = getRealItemLvl(slotTwoItemLink)
	local currentHighestLevel = 0
	local secondHighestLevel = nil
	local secondHighestLink = nil
	--Get the 2 highest items in bags (if there are 2)
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local _, _, _, _, _, _, itemLink, _, _, itemId = GetContainerItemInfo(bag, slot)
			if (itemId ~= nil and itemId ~= '') then
				local itemName, itemLink, itemQuality, itemLevel, itemMinLevel, _, _, _, itemEquipLoc = GetItemInfo(itemLink)
				--Get real item level from the visual link, GetItemInfo() returns wrong ilvl on timewalking gear and some other stuff.
				itemLevel = getRealItemLvl(itemLink)
				if (itemEquipLoc ~= nil and itemEquipLoc ~= '') then
					--If item is equippable then check the corresponding equipped slot
					local itemSlotNumber = equipmentSlotCache[itemEquipLoc]
					if (slotNumber == itemSlotNumber) then
						if (itemLevel >= currentHighestLevel) then
							secondHighestLevel = currentHighestLevel
							secondHighestLink = currentHighestLink
							secondHighestEquipLoc = currentHighestEquipLoc
							secondHighestSlot = currentHighestSlot
							secondHighestBag = currentHighestBag
							currentHighestLevel = itemLevel
							currentHighestLink = itemLink
							currentHighestEquipLoc = itemEquipLoc
							currentHighestSlot = slot
							currentHighestBag = bag
						elseif (itemLevel >= secondHighestLevel) then
							secondHighestLevel = itemLevel
							secondHighestLink = itemLink
							secondHighestEquipLoc = itemEquipLoc
							secondHighestSlot = slot
							secondHighestBag = bag
						end
					end
				end
			end
		end
	end
	--If second slot lower than first, UseContainerItem() replaces 2nd ring, otherwise even ilvl or first ring lower it equips first slot.
	if (slotTwoItemLevel < slotOneItemLevel) then
		equippedMsg = slotTwoItemLevel
		slotTwo = true
	else
		equippedMsg = slotOneItemLevel
		slotTwo = false
	end
	--If both rings in bags are higher than both rings equipped then we can just equip both and game client will decide the slots.
	if (currentHighestLevel > slotOneItemLevel and currentHighestLevel > slotTwoItemLevel
			and secondHighestLevel > slotOneItemLevel and secondHighestLevel > slotTwoItemLevel) then
		foundScannedUpgrade(msg, currentHighestBag, currentHighestSlot, currentHighestLevel, currentHighestLink, currentHighestEquipLoc, equippedMsg, slotTwo)
		foundScannedUpgrade(msg, secondHighestBag, secondHighestSlot, secondHighestLevel, secondHighestLink, secondHighestEquipLoc, equippedMsg, slotTwo)
	--Else if highest in bags is higher than only one equipped then just equip the one highest item.
	elseif (currentHighestLevel > slotOneItemLevel or currentHighestLevel > slotTwoItemLevel) then
		foundScannedUpgrade(msg, currentHighestBag, currentHighestSlot, currentHighestLevel, currentHighestLink, currentHighestEquipLoc, equippedMsg, slotTwo)	
	end
end

function foundScannedUpgrade(msg, bag, slot, itemLevel, itemLink, itemEquipLoc, equippedMsg, slotTwo)
	--If it's a slot 2 ring or trinket change the equip slot msg a bit.
	if (slotTwo == true) then
		itemLocName = _G[itemEquipLoc] .. "2"
	else
		itemLocName = _G[itemEquipLoc]
	end
	if (msg == "equip") then
		Multimod:chatMsg("|cffff8700(Equipping)|cffFFFF00 " .. _G[itemEquipLoc] .. " |cff00FF00 - Old:|cffFFFF00 " .. equippedMsg ..  
				" |cff00FF00New:|cffFFFF00 " .. itemLevel .. " " .. itemLink ..  "|cff00FF00.")
		UseContainerItem(bag, slot)
		mmScanCount = mmScanCount + 1
	else
		Multimod:chatMsg("|cffFFFF00" .. itemLocName .. " |cff00FF00 - Equipped:|cffFFFF00 " .. equippedMsg ..  
				" |cff00FF00Bag:|cffFFFF00 " .. itemLevel .. " " .. itemLink ..  "|cff00FF00.")
		mmScanCount = mmScanCount + 1
	end
end

--Get highest item in your bags that fits a certain equipment slot number.
function getHighestItemPerSlot(slotNumber)
	local highestItemLevel = 0
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local _, _, _, _, _, _, itemLink, _, _, itemId = GetContainerItemInfo(bag, slot)
			if (itemId ~= nil and itemId ~= '') then
				local itemName, itemLink, itemQuality, itemLevel, itemMinLevel, _, _, _, itemEquipLoc = GetItemInfo(itemLink)
				if (itemEquipLoc ~= nil and itemEquipLoc ~= '') then
					local equipmentSlotNumber = equipmentSlotCache[itemEquipLoc]
					--Slot of bag item matches equip slot we are checking, now check if it's the highest ilvl one in bags.
					if (slotNumber == equipmentSlotNumber and itemLevel > highestItemLevel) then
						highestItemLevel = itemLevel
						highestItemLink = itemLink
						highestBag = bag
						highestSlot = slot
						--print("Slot: " .. slotNumber .. " Slotnumber: " .. equipmentSlotNumber .. " ilvl: " .. itemLevel .. " hilvl: " .. highestItemLevel)
					end
				end
			end
		end
	end
	return highestItemLink, highestItemLevel, highestBag, highestSlot
end

--Cycle through bags and equipped items and force the game client to cache items it hasn't loaded.
--This runs roughly a minute after logging in
--Even if you wait a long time the game client still doesn't always populate the entire bag and equipped cache properly.
function Multimod:buildItemCache()
	--Cycle bags
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local itemLink = GetContainerItemLink(bag, slot)
			if (itemLink ~= nil and itemLink ~= '') then
				local slotNumber = equipmentSlotCache[itemEquipLoc]
				local equippedItemLink = GetInventoryItemLink("player", slotNumber)
				GetItemInfo(itemLink)
			end
		end
	end
	--Cycle equipped items
	for i = 1, 17 do
		local equippedItemLink = GetInventoryItemLink("player", i)
		if (equippedItemLink ~= nil and equippedItemLink ~= '') then
			GetItemInfo(equippedItemLink)
		end
	end
end

--Thanks to Personal Loot Helper for the following 2 functions that get equipment's real item level
local tooltip
local function CreateEmptyTooltip()
    local tip = CreateFrame('GameTooltip')
	local leftside = {}
	local rightside = {}
	local L, R
	for i = 1, 6 do
		L, R = tip:CreateFontString(), tip:CreateFontString()
		L:SetFontObject(GameFontNormal)
		R:SetFontObject(GameFontNormal)
		tip:AddFontStrings(L, R)
		leftside[i] = L
		rightside[i] = R
	end
	tip.leftside = leftside
	tip.rightside = rightside
	return tip
end

function getRealItemLvl(item)
	local realILVL = nil
	if item ~= nil then
		tooltip = tooltip or CreateEmptyTooltip()
		tooltip:SetOwner(UIParent, 'ANCHOR_NONE')
		tooltip:ClearLines()
		tooltip:SetHyperlink(item)
		local t = tooltip.leftside[2]:GetText()
		if t ~= nil then
--			realILVL = t:match('Item Level (%d+)')
			realILVL = t:match(_G.ITEM_LEVEL:gsub('%%d', '(%%d+)'))
		end
		-- ilvl can be in the 2nd or 3rd line dependng on the tooltip; if we didn't find it in 2nd, try 3rd
		if realILVL == nil then
			t = tooltip.leftside[3]:GetText()
			if t ~= nil then
--				realILVL = t:match('Item Level (%d+)')
				realILVL = t:match(_G.ITEM_LEVEL:gsub('%%d', '(%%d+)'))
			end
		end
		tooltip:Hide()
		-- if realILVL is still nil, we couldn't find it in the tooltip - try grabbing it from getItemInfo, even though
		-- that doesn't return upgrade levels
		if realILVL == nil then
			_, _, _, realILVL, _, _, _, _, _, _, _ = GetItemInfo(item)
		end
	end
	if realILVL == nil then
		return 0
	else		
		return tonumber(realILVL)
	end
end

-------------
--Keystones--
-------------

--Auto insert keystones when in dungeon.
function insertKeystone()
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local name=GetContainerItemLink(bag,slot)
			if (name and string.find(name,"Keystone:")) then
				UseContainerItem(bag,slot)
				Multimod:chatPrefixMsg("Inserting keystone: " .. name .. ".")
			end
		end
	end
end

--Loot handling to track multiple account keystones etc (looted keystones show up in db, loot weekly keystones while in your MB group)
local f = CreateFrame("Frame")
f:RegisterEvent("CHAT_MSG_LOOT")
f:SetScript("OnEvent", function(self, event, message)
    --Get itemlink by checking all possible matches for self loot msgs and other players loot msg.
    --Check multiple msgs first so you dont end up with item links with x2 on the end.
    --Self receive multiple loot "You receive loot: [Item]x2"
    local itemLink = message:match(LOOT_ITEM_SELF_MULTIPLE:gsub("%%s", "(.+)"))
    local whoLooted = false
    if (not itemLink) then
    	--Self loot multiple item "You receive item: [Item]x2"
    	itemLink = message:match(LOOT_ITEM_PUSHED_SELF_MULTIPLE:gsub("%%s", "(.+)"))
    	if (not itemLink) then
    		--Self receive single loot "You receive loot: [Item]"
    		itemLink = message:match(LOOT_ITEM_SELF:gsub("%%s", "(.+)"))
    		if (not itemLink) then
    	 		--Self receive single item "You receive item: [Item]"
    			itemLink = message:match(LOOT_ITEM_PUSHED_SELF:gsub("%%s", "(.+)"))
    		end
    	end
    end
    --If no matches for self loot then check other player loot msgs.
    if (not itemLink) then
    	--Other player receive multiple loot "Player receives loot: [Item]x2"
    	whoLooted, itemLink = message:match("^" .. LOOT_ITEM_MULTIPLE:gsub("%%s", "(.+)"))
    	if (not itemLink) then
    		--Other player loot multiple item "Player receives item: [Item]x2"
    		whoLooted, itemLink = message:match("^" .. LOOT_ITEM_PUSHED_MULTIPLE:gsub("%%s", "(.+)"))
    		if (not itemLink) then
    			--Other player receive single loot "Player receives loot: [Item]"
    			whoLooted, itemLink = message:match("^" .. LOOT_ITEM:gsub("%%s", "(.+)"))
    			if (not itemLink) then
    	 			--Other player receive single item "Player receives item: [Item]"
    				whoLooted, itemLink = message:match("^" .. LOOT_ITEM_PUSHED:gsub("%%s", "(.+)"))
    			end
    		end
    	end
    end
    --Insert keystones into the database.
    if itemLink and (itemLink:find("%[Keystone:") or itemLink:find("%[Mythic Keystone%]")) then
    	if (not whoLooted) then
			Multimod.data.global.keystones[Multimod.nameAndServer] = {
				keystone = itemLink,
				class = class,
			}
			--Had to add a delay here, when it's looted the bag doesn't update fast enough.
			C_Timer.After(2, function()
				updateSelfKeystone()
			end)
		else 
			--Send your keystone when anyone loots one anyway, helps keep everything in sync.
			sendKeystoneParty()
		end
	end
end)

--Insert your own keystone into db (also updates party).
function updateSelfKeystone()
	local _, class = UnitClass("player");
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
      		local tempItemLink=GetContainerItemLink(bag,slot)
      		if (tempItemLink and tempItemLink:find("%[Keystone: ")) then
				itemLink = tempItemLink
			end
		end
	end
	if (itemLink) then
		local kstime = GetServerTime()
		Multimod.data.global.keystones[Multimod.nameAndServer] = {
			keystone = itemLink,
			class = class,
			time = kstime,
		}
		sendKeystoneParty()
	end
end

--Update party members with your keystone.
function sendKeystoneParty()
	local _, class = UnitClass("player");
	local kstime = GetServerTime()
	if (not kstime or kstime == "") then
    	kstime = 0
    end
	doKeystoneTeam = true
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
      		local itemLink=GetContainerItemLink(bag,slot)
      		if (itemLink and itemLink:find("%[Keystone: ")) then
				Multimod:sendAddonChannel("PARTY", "ksupdate " .. class .. "," .. kstime .. "," .. itemLink)
				return;
			end
		end
	end
end

function receiveKeystoneParty(sender, class, kstime, itemLink)
	if (sender == UnitName("player")) then
		return
	end
	if (string.match(sender, "%-")) then
    	local key = string.gsub(sender, "-", " ")
    	local split = explode(" ", key)
		local whoLooted = split[1]
		local realm = split[2]
    	otherNameAndServer = sender .. "_" .. realm
    else
    	--No realm attached to looter name so must be our realm.
    	otherNameAndServer = sender .. "_" .. Multimod.homeServer
    end
    if (not kstime or kstime == "") then
    	kstime = GetServerTime()
    end
    if (not kstime or kstime == "") then
    	kstime = 0
    end
    Multimod.data.global.keystones[otherNameAndServer] = {
    	keystone = itemLink,
    	class = class,
    	time = tonumber(kstime),
	}
end

--Update party when someone joins.
Multimod.doKeystoneTeam = true
local f = CreateFrame("Frame")
f:RegisterEvent("GROUP_ROSTER_UPDATE")
f:SetScript('OnEvent', function(self, event, ...)
	--if (IsInRaid()) then
	--	return
	--end
	if (Multimod.doKeystoneTeam) then
		--Add a throddle so as not to spam when joining large groups.
		Multimod.doKeystoneTeam = false
		C_Timer.After(2, function()
			updateSelfKeystone() --This does self and party.
		end)
	end
end)

--Update keystone when dungeon complete.
local f = CreateFrame("Frame")
f:RegisterEvent("CHALLENGE_MODE_COMPLETED")
f:SetScript("OnEvent", function(self, event)
	C_Timer.After(2, function()
		updateSelfKeystone()
	end)
end)

--Sync keystones between alts online in party (mostly juat a debug command).
SLASH_SYNCKSCMD1 = '/syncks';
function SlashCmdList.SYNCKSCMD(msg, editBox)
	updateSelfKeystone()
	Multimod:chatPrefixMsg("Sending your Keystone info to others in party with Multimod installed.")
end

--Weekly/Daily reset, delete last weeks keystone data.
--This event only fires if you have some daily/weekly quest/event to reset but it happens to 99% of people that played during the week.
--The data deletion func also runs at logon and new keystones overwrite old, so doesn't really matter if this doesn't fire.
local f = CreateFrame("Frame")
f:RegisterEvent("CHAT_MSG_SYSTEM")
f:SetScript('OnEvent', function(self, event, msg)
	if (msg == "Daily quests are now reset!") then
		C_Timer.After(10, function()
			Multimod:wipeOldKeystoneData()
		end)
	end
end)

--Delete keystone data old than the last weekly reset time (called at logon and weekly reset event).
function Multimod:wipeOldKeystoneData()
	local lastResetTimestamp = calcWeeklyReset()
	local data = Multimod.data.global.keystones
	for key, value in pairs(data) do
		if (value.time and value.time < lastResetTimestamp) then
			Multimod.data.global.keystones[key] = nil
		end
	end
end

--Print msg if you have a keystone for the instance you entered. (This is mostly for multiboxers)
local f = CreateFrame("Frame")
f:RegisterEvent("ZONE_CHANGED_NEW_AREA")
f:SetScript('OnEvent', function(self, event, ...)
	local instanceName, instanceType, _, instanceDifficulty = GetInstanceInfo()
	--Legion Dungeons.
	if (instanceName == "Black Rook Hold" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Cathedral of Eternal Night" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Court of Stars" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Darkheart Thicket" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Eye of Azshara" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Halls of Valor" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Maw of Souls" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Neltharion's Lair" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Return to Karazhan" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Seat of the Triumvirate" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "The Arcway" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Vault of the Wardens" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	--BFA Dungeons.
	elseif (instanceName == "Siege of Boralus" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Freehold" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Shrine of the Storm" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Tol Dagor" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Waycrest Manor" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Atal'Dazar" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "The MOTHERLODE!!!" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "Temple of Sethraliss" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "The Underrot" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	elseif (instanceName == "King's Rest" and instanceDifficulty == "Mythic") then
		checkInstanceKeystone(instanceName)
	end
	
end)

function checkInstanceKeystone(dungeonName)
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			local name=GetContainerItemLink(bag,slot)
			if (name and string.find(name, "Keystone: " .. dungeonName)) then
				Multimod:chatPrefixMsg("You have a keystone for this dungeon: " .. name .. ".")
				return
			end
		end
	end
end

---------------
--Work Orders--
---------------

--Start work orders when talking to garrison/order hall npcs
local f = CreateFrame("Frame")
f:RegisterEvent("SHIPMENT_CRAFTER_OPENED")
f:SetScript("OnEvent", function(self, event, ...)
	--GarrisonCapacitiveDisplayFrame:Hide()
	--GarrisonCapacitiveDisplayFrame:Show()
	local startOrders = false
	local npcName = UnitName("target")
	local zoneName = GetZoneText()
	local subZoneName = GetSubZoneText()
	--local continent = GetCurrentMapContinent();
	if (Multimod.db.global.work_orders_orderhall) then
		if zoneName == "Dreadscar Rift" then --Warlock
			startOrders = true
		elseif zoneName == "Acherus: The Ebon Hold" then --Death Knight
			startOrders = true
		elseif zoneName == "Hall of the Guardian" then --Mage
			startOrders = true
		elseif zoneName == "Skyhold" then --Warrior
			startOrders = true
		elseif zoneName == "Trueshot Lodge" then --Hunter
			startOrders = true
		elseif zoneName == "Mardum, the Shattered Abyss" then --Demon Hunter
			startOrders = true
		elseif subZoneName == "The Hall of Shadows" or subZoneName == "Den of Thieves" then --Rogue
			startOrders = true
		elseif zoneName == "Netherlight Temple" then --Priest
			startOrders = true
		elseif zoneName == "The Maelstrom" then --Shaman
			startOrders = true
		elseif zoneName == "The Wandering Isle" then --Monk
			startOrders = true
		elseif zoneName == "Light's Hope Chapel" then --Paladin
			startOrders = true
		elseif zoneName == "The Dreamgrove" then --Druid
			startOrders = true
		elseif zoneName == "Wind's Redemption" then --Boralus alliance ship.
			startOrders = true
		elseif zoneName == "The Banshee's Wail" then --Horde ship.
			startOrders = true
		end
	end
	--if (Multimod.db.global.work_orders_garrison and continent == 7) then
		--We check world map ID to get the continent number for Draenor instead of zone name for Garrison work orders.
		--Every single building type in Garrison has it's own zone name for some reason.
		--World map id == 7 is enough since the only work orders on Draenor are in the Garrison anyway.
	--	startOrders = true
	--end
	if (Multimod.db.global.work_orders_nomi and npcName == "Nomi") then
		startOrders = true
	end
	if (startOrders) then
		--We need a 1 second delay before doing anything, the game client doesn't get the frame info instantly on load.
		--On rare occasions it still doesn't get the info even with a 1 second delay.
		C_Timer.After(1, function()
			startAllWorkOrders()
		end)
	end
end)

function startAllWorkOrders()
	local buttonState = GarrisonCapacitiveDisplayFrame.CreateAllWorkOrdersButton:GetButtonState()
	if (buttonState == "NORMAL") then
		local workOrderType = GarrisonCapacitiveDisplayFrame.CapacitiveDisplay.ShipmentIconFrame.ShipmentName:GetText()
		local workOrdersAvailable = GarrisonCapacitiveDisplayFrame.CapacitiveDisplay.ShipmentIconFrame.ShipmentsAvailable:GetText()
		if (workOrderType ~= nil and workOrderType ~= '' and workOrdersAvailable ~= nil and workOrdersAvailable ~= '') then
			local workOrderCount = {}
			workOrderCount[1], workOrderCount[2] = workOrdersAvailable:match("(%w+)(.+)")
			GarrisonCapacitiveDisplayFrame.CreateAllWorkOrdersButton:Click()
			Multimod:chatPrefixMsg("Started|cffFFFF00 " .. workOrderCount[1] .. " |cff00FF00work orders for|cffFFFF00 " .. 
					workOrderType .. "|cff00FF00.")
		else
			GarrisonCapacitiveDisplayFrame.CreateAllWorkOrdersButton:Click()
			Multimod:chatPrefixMsg("Starting all work orders.")
		end
	end
end

--------------------------------------------------------------------
--Flash WoW client when minimized on certain events. (invites etc)--
--------------------------------------------------------------------

local lastBrawlMsg = 0
local f = CreateFrame("Frame")
f:RegisterEvent("PARTY_INVITE_REQUEST") --Group invite.
--f:RegisterEvent("PARTY_INVITE_XREALM") --I think this is not used anymore, putting here incase I missed something though.
f:RegisterEvent("READY_CHECK") --Ready check.
f:RegisterEvent("DUEL_REQUESTED") --Duel requested.
f:RegisterEvent("CONFIRM_SUMMON") --Summon popup.
f:RegisterEvent("TRADE_SHOW") --Trade started.
f:RegisterEvent("TRADE_ACCEPT_UPDATE")	--Trade accept button clicked by partner.
f:RegisterEvent("CHAT_MSG_MONSTER_YELL") --Brawlers invite.
f:RegisterEvent("UNIT_AURA") --Brawlers first in line to fight.
f:RegisterEvent("PLAYER_REGEN_DISABLED") --Combat started.
f:RegisterEvent("MIRROR_TIMER_START") --Fatigue started.
--f:RegisterEvent("VIGNETTE_ADDED") --Rare spawned.
f:RegisterEvent("CHAT_MSG_WHISPER") --Whisper.
f:RegisterEvent("CHAT_MSG_BN_WHISPER") --Bnet Whisper.
f:RegisterEvent("PET_BATTLE_PVP_DUEL_REQUESTED") --Pet duel.
f:RegisterEvent("BN_FRIEND_INVITE_ADDED") --Bnet friend request.
--f:RegisterEvent("GMRESPONSE_RECEIVED") --GM answers ticket.
f:SetScript('OnEvent', function(self, event, ...)
	if (Multimod.db.global.flash_wow_client == false) then
		return
	end
	local arg1, arg2, arg3, arg4, arg5 = ...
	if (event == "CHAT_MSG_MONSTER_YELL" and arg5) then
		local _, _, _, _, _, _, _, instanceMapID = GetInstanceInfo()
		--Horde brawlers = 1043 // Alliance brawlers = 369
		if (arg5 == UnitName("player") and (instanceMapID == 1043 or instanceMapID == 369)) then
			Multimod:startFlash()
		end
	elseif (event == "UNIT_AURA") then
		local _, _, _, _, _, _, _, instanceMapID = GetInstanceInfo()
		if (instanceMapID == 1043 or instanceMapID == 369) then
			local _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, queuePosition = UnitBuff("player", "Queued for Brawl")
			if (queuePosition == 1 and (lastBrawlMsg < (GetServerTime() - 120))) then
				Multimod:chatPrefixMsg("You are next in line for a brawl.")
				lastBrawlMsg = GetServerTime()
				Multimod:startFlash()
			end
		end
	else
		--All other registered events don't need additional checking, just flash.
		Multimod:startFlash()
	end
end)

--All kinds of dungeon/raid/pvp/etc invites.
hooksecurefunc("StaticPopupSpecial_Show", function(...)
	if (Multimod.db.global.flash_wow_client == false) then
		return
	end
   Multimod:startFlash()
end)

--Flight paths.
local doCheckLeaveFlghtPath = false
hooksecurefunc("TakeTaxiNode", function(...)
	if (Multimod.db.global.flash_wow_client == false) then
		return
	end
	doCheckLeaveFlghtPath = true
    --Give it a few seconds to get on the taxi.
    C_Timer.After(5, function()
		checkLeaveFlghtPath()
	end)
end)

--Loop this func till flight path is left.
function checkLeaveFlghtPath()
    local isOnFlightPath = UnitOnTaxi("player")
    if (not isOnFlightPath) then
    	doCheckLeaveFlghtPath = false
    	 Multimod:startFlash()
    end
    if (doCheckLeaveFlghtPath) then
    	C_Timer.After(2, function()
			checkLeaveFlghtPath()
		end)
	end
end

local lastFlash = 0
function Multimod:startFlash()
	if (lastFlash < (GetServerTime() - 4)) then
		FlashClientIcon()
		lastFlash = GetServerTime()
	end
end

------------------------------
--Skip Cut scenes and Movies--
------------------------------

local f = CreateFrame("Frame")
f:RegisterEvent("CINEMATIC_START")
f:RegisterEvent("PLAY_MOVIE")
f:SetScript('OnEvent', function(self, event, ...)
	if (Multimod.db.global.skip_cut_scenes == false) then
		return
	end
	--This is to prevent cut scene skip during the quest "The Big Boom" in Vol'dun, it bugs if the cut scene skipped.
	if (C_Map.GetBestMapForUnit("player") == 864 and (UnitInVehicle("player") or UnitControllingVehicle("player"))) then
		return
	end
	if (event == "CINEMATIC_START") then
		CinematicFrame_CancelCinematic()
		Multimod:chatPrefixMsg("Cut scene skipped.")
	elseif (event == "PLAY_MOVIE") then
		GameMovieFinished()
		MovieFrame:Hide()
		Multimod:chatPrefixMsg("Ingame movie skipped.")
	end
end)

-----------------------------
--Auto accept group invites--
-----------------------------

local f = CreateFrame("Frame")
f:RegisterEvent("PARTY_INVITE_REQUEST") --Group invite.
--f:RegisterEvent("PARTY_INVITE_XREALM") --I think this is not used anymore, putting here incase I missed something though.
f:SetScript('OnEvent', function(self, event, player)
	--Strip realm from name.
	local data = explode("-", player)
	local player = data[1]
	if (not player or IsInGroup() or IsInRaid()) then
		return
	end
	if (Multimod.db.global.autoaccept_allinvites) then
		doAcceptGroup(player)
		return
	end
	if (Multimod.db.global.autoaccept_guildinvites and IsInGuild()) then
		local guildCount = GetNumGuildMembers()
		for i = 1, guildCount do
			local guildPlayer = GetGuildRosterInfo(i)
			local data = explode("-", player)
			local guildPlayer = data[1]
			if (guildPlayer and guildPlayer == player) then
				doAcceptGroup(player)
				return
			end
		end
	end
	if (Multimod.db.global.autoaccept_friendinvites) then
		local totalFriends, onlineFriends = GetNumFriends()
		if (onlineFriends ~= nil) then
			for i = 1, onlineFriends do
				local name = GetFriendInfo(i);
				if (name and name == player) then
					doAcceptGroup(player)
					return
				end
			end
		end
	end
	if (Multimod.db.global.autoaccept_bnetfriendinvites) then
		local totalBNFriends, onlineBNFriends = BNGetNumFriends()
		if (onlineBNFriends ~= nil) then
			for friendIndex = 1, onlineBNFriends do
				if ((BNGetFriendGameAccountInfo(friendIndex, 1) ~= nil) and (BNGetFriendGameAccountInfo(friendIndex, 1) ~= '')) then
					local i = 1
					local _, toonName = BNGetFriendGameAccountInfo(friendIndex, i)
					--Cycle all chars online for this bnet friend incase multiple accounts.
    				while toonName do
						if (toonName and toonName == player) then
							doAcceptGroup(player)
							return
						end
      					i = i + 1
      					_, toonName, _, realmName, _, faction, _, class, _, _, level = BNGetFriendGameAccountInfo(friendIndex, i)
      				end
    			end
			end
		end
	end
end)

function doAcceptGroup(player)
	--Don't join if queued for something.
	pvpQueue = GetBattlefieldStatus(1)
	worldPvpQueue = GetWorldPVPQueueStatus(1)
	if (Multimod.db.global.autoaccept_notwhilequeued and (GetLFGQueueStats(LE_LFG_CATEGORY_LFD) or GetLFGQueueStats(LE_LFG_CATEGORY_LFR)
			or GetLFGQueueStats(LE_LFG_CATEGORY_RF) or GetLFGQueueStats(LE_LFG_CATEGORY_SCENARIO)
			or GetLFGQueueStats(LE_LFG_CATEGORY_FLEXRAID) or GetBattlefieldEstimatedWaitTime(1) > 0
			or GetLFGQueueStats(LE_LFG_CATEGORY_WORLDPVP) or pvpQueue == "queued"
			or (worldPvpQueue ~= "none" and worldPvpQueue ~= nil))) then
		--Delay msg for 1 second to see if another addon forced them into group.
		C_Timer.After(1, function()
			sendAcceptGroupMsg(player)
		end)
		return	
	end
	AcceptGroup()
	--Hide invite popup.
	for i = 1, STATICPOPUP_NUMDIALOGS do
		local dialogType = _G["StaticPopup"..i].which
		if (dialogType == "PARTY_INVITE" or dialogType == "PARTY_INVITE_XREALM") then
			_G["StaticPopup"..i].inviteAccepted = 1
			StaticPopup_Hide(dialogType)
		end
	end
end

--Delay msg for 1 second to see if another addon forced them into group.
function sendAcceptGroupMsg(player)
	if (IsInGroup() or IsInRaid()) then
		Multimod:chatPrefixMsg("Skipped joining group invite from " .. player .. " because you are in a queue" ..
				" (It looks like another addon made you join group though?).")
	else
		Multimod:chatPrefixMsg("Skipped joining group invite from " .. player .. " because you are in a queue.")
	end
end

--Heirloom list.
local mmLoomList = {
	--Helm
	"Burnished Helm of Might",
	"Polished Helm of Valor",
	"Pristine Lightforge Helm",
	"Mystical Coif of Elements",
	"Tarnished Raging Berserker%'s",
	"Preened Tribal War Feathers",
	"Stained Shadowcraft Cap",
	"Tattered Dreadmist Mask",
	--Shoulder
	"Burnished Pauldrons of Might",
	"Polished Spaulders of Valor",
	"Pristine Lightforge Spaulders",
	"Strengthened Stockade Pauldrons",
	"Aged Pauldrons of The Five Thunders",
	"Champion Herod%'s Shoulder",
	"Mystical Pauldrons of Elements",
	"Prized Beastmaster%'s Mantle",
	"Exceptional Stormshroud Shoulders",
	"Lasting Feralheart Spaulders",
	"Preened Ironfeather Shoulders",
	"Stained Shadowcraft Spaulders",
	"Exquisite Sunderseer Mantle",
	"Tattered Dreadmist Mantle",
	--Back
	"Ancient Bloodmoon Cloak",
	"Inherited Cape of the Black Baron",
	"Ripped Sandstorm Cloak",
	"Worn Stoneskin Gargoyle Cape",
	--Chest
	"Burnished Breastplate of Might",
	"Polished Breastplate of Valor",
	"Pristine Lightforge Breastplate",
	"Champion%'s Deathdealer Breastplate",
	"Mystical Vest of Elements",
	"Preened Ironfeather Breastplate",
	"Stained Shadowcraft Tunic",
	"Tattered Dreadmist Robe",
	--Legs
	"Burnished Legplates of Might",
	"Polished Legplates of Valor",
	"Pristine Lightforge Legplates",
	"Mystical Kilt of Elements",
	"Tarnished Leggings of Destruction",
	"Preened Wildfeather Leggings",
	"Stained Shadowcraft Pants",
	"Tattered Dreadmist Leggings",
	--Weapons
	"Bloodied Arcanite Reaper",
	"Charmed Ancient Bone Bow",
	"Upgraded Dwarven Hand Cannon",
	"Devout Aurastone Hammer",
	"The Blessed Hammer of Grace",
	"Venerable Mass of McGowan",
	"Repurposed Lava Dredger",
	"Eagletalon Spear",
	"Spear of Rethu",
	"Bloodsoaked Skullforge Reaver",
	"Venerable Dal%'Rend%'s Sacred Charge",
	"Battleworn Thrash Blade",
	"Reforged Truesilver Champion",
	"Burnished Warden Staff",
	"Dignified Headmaster%'s Charge",
	"Grand Staff of Jordan",
	"Brawler%'s Razor Claws",
	"Balanced Heartseeker",
	"Sharpened Scarlet Kris",
	"Flamescarred Draconian Deflector",
	"Weathered Observe%'s Shield",
	"Musty Tome of the Lost",
	--Trinkets, Rings, & Necklaces
	"Admiral Taylor%'s Loyalty Ring",
	"Captain Sander%'s Returned Band",
	"Defending Champion",
	"Discerning Eye of the Beast",
	"Dread Pirate Ring",
	"Eternal Amulet of the Redeemed",
	"Eternal Emberfury Talisman",
	"Eternal Horizon Choker",
	"Eternal Talisman of Evasion",
	"Eternal Will of the Martyr",
	"Eternal Woven Ivy Necklace",
	"Man%'ari Training Amulet",
	"Returning Champion",
	"Signet of the Third Fleet",
	"Swift Hand of Justice",
	"Touch of the Void",
	"Inherited Insignia of the Alliance",
	"Inherited Insignia of the Horde",
	"Inherited Mark of Tyranny",
	--Seige of Orgrimmar Heirlooms
	"Hellscream%'s Tome of Destruction",
	"Hellscream%'s Warmace",
	"Hellscream%'s Pig Sticker",
	"Hellscream%'s Razor",
	"Hellscream%'s Doomblade",
	"Hellscream%'s War Staff",
	"Hellscream%'s Barrier",
	"Hellscream%'s Cleaver",
	"Hellscream%'s Shield Wall",
	"Hellscream%'s Warbow",
	"Hellscream%'s Decapitator",
	--WoD Dungeon Trinkets
	"Judgment of the Naaru",
	"Gronntooth War Horn",
	"Orb of Voidsight",
	"Infallible Tracking Charm",
	"Purified Shard of the Third Moon",
}
 
--Delete heirlooms "/deletelooms".
SLASH_DESTROYLOOMSCMD1, SLASH_DESTROYLOOMSCMD2 = '/destroylooms', '/deletelooms';
function SlashCmdList.DESTROYLOOMSCMD(msg, editBox)
	if (Multimod.startDelay > time() and msg ~= "lowSpam") then
		Multimod:chatPrefixMsg("Please wait " .. Multimod.startDelaySeconds .. " seconds after logging in so the item cache can load properly. (" .. 
				(Multimod.startDelay - time()) .. " seconds left)")
		return
	end
	if (Multimod.db.global.delete_looms_enchanted) then
		Multimod:chatPrefixMsg("Destroying all unequipped Heirlooms...")
	else
		Multimod:chatPrefixMsg("Destroying all unequipped Heirlooms...")
		Multimod:chatMsg("(Skipping enchanted items is set in options)")
	end
	local count = 0
	for bag = 0, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
    		for i, v in ipairs(mmLoomList) do
      			local name=GetContainerItemLink(bag,slot)
				if (name and string.find(name, v) and name ~= nill and name ~= '') then
					local itemId, enchantId, gem1, gem2, gem3 = string.match(name, "item:?(%d*):?(%d*):?(%d*):?(%d*):?(%d*)")
    				if (enchantId and enchantId ~= '' and tonumber(enchantId) >= 1 and not Multimod.db.global.delete_looms_enchanted) then
    					Multimod:chatMsg("Skipping " .. name .. " |cff00FF00because it's enchanted.")
    				else
						PickupContainerItem(bag,slot)
						DeleteCursorItem()
						Multimod:chatMsg("Destroyed: " .. name .. "|cff00FF00.")
						count = count + 1
					end
				end
			end
		end
	end
	if (msg ~= "lowSpam") then
		Multimod:chatMsg("Done, " .. count .. " items found and destroyed.")
	end
end

--------------------------------------------------------------------------
--Move the TalkingHeadFrame that appeared in Legion for world quests etc--
--------------------------------------------------------------------------

--This runs when the addon loads, create the frame, blizzard doesn't usually load it till the first use.
function Multimod:talkingHeadLoad()
	--Got this from UIParent.lua to load the frame.
	if (not TalkingHeadFrame) then
			TalkingHead_LoadUI();
	end
	TalkingHeadFrame:SetMovable(true)
	TalkingHeadFrame:EnableMouse(true)
	--Adding it to special frames makes it closable with ESC.
	tinsert(UISpecialFrames, "TalkingHeadFrame")
	TalkingHeadFrame:SetScript("OnMouseDown", function(self, button)
		if IsModifierKeyDown() then
			if button == "LeftButton" and not self.isMoving then
				self:StartMoving();
				self.isMoving = true;
			end
		end
	end)
	TalkingHeadFrame:SetScript("OnMouseUp", function(self, button)
		if button == "LeftButton" and self.isMoving then
			self:StopMovingOrSizing();
			self.isMoving = false;
			Multimod.db.global.talking_head_point, _, Multimod.db.global.talking_head_relativePoint, 
					Multimod.db.global.talking_head_x, Multimod.db.global.talking_head_y = self:GetPoint()
		end
	end)
	TalkingHeadFrame:SetScript("OnHide", function(self)
		if (self.isMoving) then
			self:StopMovingOrSizing();
			self.isMoving = false;
			Multimod.db.global.talking_head_point, _, Multimod.db.global.talking_head_relativePoint, 
					Multimod.db.global.talking_head_x, Multimod.db.global.talking_head_y = self:GetPoint()
		end
	end)
	--Set our custom size at load time.
	if (Multimod.db.global.talking_head_size) then
		TalkingHeadFrame:SetScale(Multimod.db.global.talking_head_size)
	end
	--Set position (Blizzard saves this frame position also but it seems to reset sometimes).
	if (Multimod.db.global.talking_head_point and Multimod.db.global.talking_head_relativePoint and 
			Multimod.db.global.talking_head_x and Multimod.db.global.talking_head_y) then
		TalkingHeadFrame.ignoreFramePositionManager = true
		TalkingHeadFrame:ClearAllPoints()
		TalkingHeadFrame:SetPoint(Multimod.db.global.talking_head_point, nil, Multimod.db.global.talking_head_relativePoint, 
				Multimod.db.global.talking_head_x, Multimod.db.global.talking_head_y)
		TalkingHeadFrame:SetUserPlaced(false)
	end
end

--Run a custom talking head for config so people can drag it, runs when button clicked in options.
--This one never disappears till X is clicked.
--This is a copy of blizzards default func with a few changes.
function Multimod:Custom_TalkingHeadFrame_PlayCurrent()
	if (not TalkingHeadFrame) then
			Multimod:talkingHeadLoad()
	end
	if (TalkingHeadFrame:IsShown()) then
		TalkingHeadFrame:Hide()
		return
	end
	Multimod:chatPrefixMsg("Hold shift down to drag Talking Head frame.")
	local frame = TalkingHeadFrame;
	local model = frame.MainFrame.Model;

	if( frame.finishTimer ) then
		frame.finishTimer:Cancel();
		frame.finishTimer = nil;
	end
	if ( frame.voHandle ) then
		StopSound(frame.voHandle);
		frame.voHandle = nil;
	end
	local currentDisplayInfo = model:GetDisplayInfo();
	--local displayInfo, cameraID, vo, duration, lineNumber, numLines, name, text, isNewTalkingHead = C_TalkingHead.GetCurrentLineInfo();
	local displayInfo = 73318
	local cameraID = 129
	local vo = 57231
	local duration = 14.207000732422
	local lineNumber = 0
	local numLines = 1
	local name = "MultiMod Configuration"
	local text = "Hold Shift down to drag the frame where you want it, click X when done."
	local isNewTalkingHead = true
	local textFormatted = string.format(text);
	if ( displayInfo and displayInfo ~= 0 ) then
		frame:Show();
		if ( currentDisplayInfo ~= displayInfo ) then
			model.uiCameraID = cameraID;
			model:SetDisplayInfo(displayInfo);
		else
			if ( model.uiCameraID ~= cameraID ) then
				model.uiCameraID = cameraID;
				Model_ApplyUICamera(model, model.uiCameraID);
			end
			TalkingHeadFrame_SetupAnimations(model);
		end

		if ( isNewTalkingHead ) then
			TalkingHeadFrame_Reset(frame, textFormatted, name);
			TalkingHeadFrame_FadeinFrames();
		else
			if ( name ~= frame.NameFrame.Name:GetText() ) then
				-- Fade out the old name and fade in the new name
				frame.NameFrame.Fadeout:Play();
				C_Timer.After(0.25, function()
					frame.NameFrame.Name:SetText(name);
				end);
				C_Timer.After(0.5, function()
					frame.NameFrame.Fadein:Play();
				end);

				frame.MainFrame.TalkingHeadsInAnim:Play();
			end
			if ( textFormatted ~= frame.TextFrame.Text:GetText() ) then
				-- Fade out the old text and fade in the new text
				frame.TextFrame.Fadeout:Play();
				C_Timer.After(0.25, function()
					frame.TextFrame.Text:SetText(textFormatted);
				end);
				C_Timer.After(0.5, function()
					frame.TextFrame.Fadein:Play();
				end);
			end
		end
		local success, voHandle = PlaySound(vo, "Talking Head", true, true);
		if ( success ) then
			frame.voHandle = voHandle;
		end
	end
end

--Make InterfaceOptionsFrame moveable.
function Multimod:setMoveableInterfaceOptions()
	InterfaceOptionsFrame:SetMovable(true)
	InterfaceOptionsFrame:EnableMouse(true)
	InterfaceOptionsFrame:HookScript("OnMouseDown", function(self, button)
		if (not self.isMoving and IsShiftKeyDown()) then
			self:StartMoving();
			self.isMoving = true;
		end
	end)
	InterfaceOptionsFrame:HookScript("OnMouseUp", function(self, button)
		if (self.isMoving) then
			self:StopMovingOrSizing();
			self.isMoving = false;
		end
	end)
	InterfaceOptionsFrame:HookScript("OnHide", function(self)
		if (self.isMoving) then
			self:StopMovingOrSizing();
			self.isMoving = false;
		end
	end)
end

--Add quest count back to the quest log frame.
multimodQuestCount = CreateFrame("Frame", "multimodQuestCount", QuestScrollFrame)
multimodQuestCount.fs = multimodQuestCount:CreateFontString("multimodQuestCountFS", "ARTWORK")
multimodQuestCount.fs:SetFont("Interface\\AddOns\\Multimod\\Media\\Segoe UI.ttf", 13)
multimodQuestCount.fs:SetPoint("TOP", "QuestScrollFrame", -40, 21)
multimodQuestCount:RegisterEvent("QUEST_LOG_UPDATE")
multimodQuestCount:SetScript('OnEvent', function(self, event, msg)
	if (Multimod.db.global.quest_log_count) then
		local _, numQuests = GetNumQuestLogEntries();
		multimodQuestCount.fs:SetText("|cffFFFF00Quests:  (" .. numQuests .. "/25)")
	end
end)
multimodQuestCount:SetScript('OnLoad', function(self, event, msg)
	if (Multimod.db.global.quest_log_count) then
		local _, numQuests = GetNumQuestLogEntries();
		multimodQuestCount.fs:SetText("|cffFFFF00Quests:  (" .. numQuests .. "/25)")
	end
end)

---------------------------------------
--Move/Resize the Quest Tracker Frame--
---------------------------------------

--[[function Multimod:questTrackerLoad()
	ObjectiveTrackerFrame:SetMovable(true)
	ObjectiveTrackerFrame:SetScript("OnMouseDown", function(self, button)
		if Multimod.db.global.quest_tracker_move and IsModifierKeyDown() then
			if button == "LeftButton" and not self.isMoving then
				self:StartMoving();
				self.isMoving = true;
			end
		end
	end)
	ObjectiveTrackerFrame:SetScript("OnMouseUp", function(self, button)
		if Multimod.db.global.quest_tracker_move and button == "LeftButton" and self.isMoving then
			self:StopMovingOrSizing();
			self.isMoving = false;
			Multimod.db.global.quest_tracker_point, _, Multimod.db.global.quest_tracker_relativePoint, 
					Multimod.db.global.quest_tracker_x, Multimod.db.global.quest_tracker_y = self:GetPoint()
			Multimod.db.global.quest_tracker_height = ObjectiveTrackerFrame:GetHeight()
		end
	end)
	ObjectiveTrackerFrame:SetScript("OnHide", function(self)
		if Multimod.db.global.quest_tracker_move and self.isMoving then
			self:StopMovingOrSizing();
			self.isMoving = false;
			Multimod.db.global.quest_tracker_point, _, Multimod.db.global.quest_tracker_relativePoint, 
					Multimod.db.global.quest_tracker_x, Multimod.db.global.quest_tracker_y = self:GetPoint()
			Multimod.db.global.quest_tracker_height = ObjectiveTrackerFrame:GetHeight()
		end
	end)
end

function Multimod:setQuestTracker()
	--Set our custom size at load time.
	if (Multimod.db.global.quest_tracker_size) then
		ObjectiveTrackerFrame:SetScale(Multimod.db.global.quest_tracker_size)
	end
	if (Multimod.db.global.quest_tracker_point and Multimod.db.global.quest_tracker_relativePoint and 
			Multimod.db.global.quest_tracker_x and Multimod.db.global.quest_tracker_y) then
		ObjectiveTrackerFrame:ClearAllPoints()
		ObjectiveTrackerFrame:SetPoint(Multimod.db.global.quest_tracker_point, Minimap, Multimod.db.global.quest_tracker_relativePoint, 
				Multimod.db.global.quest_tracker_x, Multimod.db.global.quest_tracker_y)
		ObjectiveTrackerFrame:SetHeight(Multimod.db.global.quest_tracker_height)
		ObjectiveTrackerFrame:SetUserPlaced(false)
	end
end--]]

------------------------------------------------
--Display in chat when you gain an honor level--
------------------------------------------------

local f = CreateFrame("Frame")
f:RegisterEvent("HONOR_LEVEL_UPDATE")
f:SetScript("OnEvent", function(self, event)
	local newLevel = UnitHonorLevel("player")
	Multimod:chatMsg("New Honor level gained|cffFFFF00 " .. UnitHonorLevel("player") .. Multimod.chatColor .. "!")
end)