---------------------------------------------------------
--Tools that certain functions in other modules require--
---------------------------------------------------------

--GetServerTime() sets the daily reset time including daylight savings changes.
--So all we need is the right day of the week for each regions reset in GMT.
function calcWeeklyReset()
	--Run this each logon to catch any daylight savings time changes during the week.
	local serverTime = GetServerTime()
	if (not serverTime) then
		--If for some reason get server time fails do nothing and try again later.
		return
	end
	local timeOfDay = serverTime + GetQuestResetTime()
	local regionName = mmLibRealmInfo:GetCurrentRegion()
	--If lib fails use wow client api (less reliable since EU client can play on US etc).
	if (not regionName) then
		local regionArray = {"US", "KR", "EU", "TW", "CN"}
		regionName = regionArray[GetCurrentRegion()]
	end
	local regionDays = {
    	["US"] = "Tuesday",
    	["EU"] = "Wednesday",
    	["KR"] = "Thursday",
    	["TW"] = "Thursday",
    	["CN"] = "Thursday",
	}
	local resetDay = regionDays[regionName]
	--If all fails just default to Tuesday.
	if (not resetDay) then
		resetDay = "Tuesday"
	end
	local dayString = date("!%A", timeOfDay)
	local nextResetTimestamp, nextResetTimestamp
	if (dayString == resetDay) then
		--Current day is reset day.
		nextResetTimestamp = timeOfDay
	else
		--If not current day then loop till we find it.
		for i=1, 7 do
			timeOfDay = timeOfDay + 86400
			dayString = date("!%A", timeOfDay)
			if (dayString == resetDay) then
				--We're at reset day GMT in the loop, set the time.
				nextResetTimestamp = timeOfDay
			end
		end
	end
	local lastResetTimestamp = nextResetTimestamp - 604800
	--print("Next reset: " .. date("!%c", nextResetTimestamp))
	--print("Last reset: " .. date("!%c", lastResetTimestamp))
	return lastResetTimestamp
end

--Credit: http://richard.warburton.it
--Modified this slightly to add count.
function explode(div, str, count)
	if (div == '') then return false end
	local pos,arr = 0,{}
	-- for each divider found
	local index = 0
	for st,sp in function() return string.find(str,div,pos,true) end do
		index = index + 1
 		table.insert(arr,string.sub(str,pos,st-1)) -- Attach chars left of current divider
		pos = sp + 1 -- Jump past current divider
		if (count and index == count) then
			table.insert(arr,string.sub(str,pos)) -- Attach chars right of last divider
			return arr
		end
	end
	table.insert(arr,string.sub(str,pos)) -- Attach chars right of last divider
	return arr
end

--Remove realm from name.
function stripRealm(name)
	args = explode(name, "-")
	if (args) then
		return args[1]
	else
		return false
	end
end

--Credit: http://www.lua.org/pil/19.3.html
--Same as ipairs, but works with associative arrays to iterate in alphabetical order.
function pairsByKeys(t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
		table.sort(a, f)
		local i = 0      -- iterator variable
		local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
			else return a[i], t[a[i]]
		end
	end
	return iter
end

--Credit: http://lua-users.org/wiki/FormattingNumbers
--Inserts commas into number values. 1000000 -> 1,000,000
function comma_value(amount)
	local formatted = amount
	while true do  
		formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
		if (k == 0) then
			break
		end
	end
	return formatted
end

--Keep a friends list cache of online players.
--Logon events will trigger when you load into the game as FRIENDLIST_UPDATE/BN_FRIEND_INFO_CHANGED are called.
onlineFriendCache = {}
local f = CreateFrame("Frame")
f:RegisterEvent("FRIENDLIST_UPDATE")
f:RegisterEvent("BN_FRIEND_INFO_CHANGED")
f:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED")
f:SetScript('OnEvent', function(self, event, ...)
	friendCache(event)
end)
function friendCache(event)
	local tempFriendCache = {}
	local currentFaction = UnitFactionGroup("player")
	--Friends list.
	local totalFriends, onlineFriends = GetNumFriends()
	if (onlineFriends ~= nil) then
		for i = 1, onlineFriends do
			local name, level, class, area, connected, status, note = GetFriendInfo(i);
			if (name ~= nil and name ~= '') then
				--If not a connectd realm then add -Realmname
				if (string.match(name, "%-") == nil) then
					name = name .. "-" .. _G.GetRealmName()
				end
				tempFriendCache[name] = {
					faction = currentFaction,
					class = class,
					level = level,
					realm = _G.GetRealmName(),
					type = "friendlist",
				}
			end
		end
	end
	--Battle.net friends list.
	local totalBNFriends, onlineBNFriends = BNGetNumFriends()
	if (onlineBNFriends ~= nil) then
		for friendIndex = 1, onlineBNFriends do
			if ((BNGetFriendGameAccountInfo(friendIndex, 1) ~= nil) and (BNGetFriendGameAccountInfo(friendIndex, 1) ~= '')) then
				local i = 1
				local _, toonName, _, realmName, _, faction, _, class, _, _, level = BNGetFriendGameAccountInfo(friendIndex, i)
				local _, _, _, _, _, presenceID = BNGetFriendInfo(friendIndex)
				--Cycle all chars online for this bnet friend incase multiple wow licences.
    			while toonName do
					if (toonName ~= nil and toonName ~= '') then
						toonName = toonName .. "-" .. realmName
						tempFriendCache[toonName] = {
							faction = currentFaction,
							class = class,
							level = level,
							realm = realmName,
							type = "bnet",
							presenceID = presenceID,
						}
					end
      				i = i + 1
      				_, toonName, _, realmName, _, faction, _, class, _, _, level = BNGetFriendGameAccountInfo(friendIndex, i)
      			end
    		end
		end
	end
	for name, v in pairs(tempFriendCache) do
		if (not onlineFriendCache[name]) then
			onlineFriendCache[name] = tempFriendCache[name]
			--Character logged ON functions here.
			if (v['faction'] == currentFaction) then
				--Same faction logons here.
				if (v['type'] == "bnet" and BNFeaturesEnabledAndConnected()) then
    				BNSendGameData(v['presenceID'], Multimod.commPrefix, "mmmap " .. mmNameAndServer .. ",addoncheck")
    			elseif (v['type'] == "friendlist") then
    				Multimod:sendAddonChannel("WHISPER", "mmmap " .. mmNameAndServer .. ",addoncheck", name)
    			end
    			--3 second delay, if they reply to addon check in those 3 seconds then we don't need to send it twice.
    			--C_Timer.After(3, function()
				--	createMapCache(true)
				--end)
			end
			--Both faction logons here.
			
		end
	end
	for name, v in pairs(onlineFriendCache) do
		if (not tempFriendCache[name]) then
			onlineFriendCache[name] = nil
			--Character logged OFF functions here.
			if (v['faction'] == currentFaction) then
				--Same faction logoffs here.
				removeMapMarker(name)
				remMmMapCache(name)
			end
			--Both faction logoffs here.
			
		end
	end
end

--Return an array of friends and bnet friends character names online.
--Requires above friendCache() function.
function getFriendsOnline(string)
	--Return format: onlineFriendCache[name] = {faction, class, level, realm, type}
	return onlineFriendCache
end

--Return true if player is online either friends list or battle.net. (ingame character not bnet account)
--Requires above friendCache() function.
function isFriendOrBnet(player, queryServer)
	if (queryServer) then
		ShowFriends()
	end
	if (onlineFriendCache[player]) then
		return true
	end
end

--Gets a time string for the period between now and seconds. (seconds being a timestamp in the future)
--If countOnly is true then we're just feeding it an amount of seconds to get string from.
function getTimeString(seconds, countOnly)
	local timecalc = 0
	if (countOnly) then
		timecalc = seconds
	else
		timecalc = seconds - time()
	end
	local y = math.floor((timecalc / (86400*365)))
	local mo = math.floor(((timecalc % (86400*365)) / (86400*31)))
	local w = math.floor((timecalc % (86400*31)) / (86400*7))
	local d = math.floor((timecalc % (86400*7)) / 86400)
	local h = math.floor((timecalc % 86400) / 3600)
	local m = math.floor((timecalc % 3600) / 60)
	local s = math.floor((timecalc % 60))
	if (y == 1) then
		return y .. " year, " .. mo .. " months";
	end
	if (y > 1)  then
		return y .. " years, " .. mo .. " months";
	end
	if (mo == 1) then
		return mo .. " month, " .. w .. " weeks";
	end
	if (mo > 1) then
		return mo .. " months, " .. w .. " weeks";
	end
	if (w == 1) then
		return w .. " week, " .. d .. " days";
	end
	if (w > 1) then
		return w .. " weeks, " .. d .. " days";
	end
	if (d == 1) then
		return d .. " day, " .. h .. " hours";
	end
	if (d > 1) then
		return d .. " days, " .. h .. " hours";
	end
	if (h == 1) then
		return h .. " hour, " .. m .. " minutes";
	end
	if (h > 1) then
		return h .. " hours, " .. m .. " minutes";
	end
	if (m == 1) then
		return m .. " minute " .. s .. " seconds";
	end
	if (m > 1) then
		return m .. " minutes " .. s .. " seconds";
	end
	--If no matches it must be seconds only.
	return s .. " seconds";
end

--Get the time until next world quest rotation.
function nextWqCycle()
	serverTime = GetServerTime()
	if (not serverTime) then
		--If for some reason get server time fails do nothing and try again later.
		return
	end
	timeOfDay = serverTime + GetQuestResetTime()
	secondsToDailyReset = GetQuestResetTime()
	if (secondsToDailyReset <= 86400 and secondsToDailyReset >= 64800) then
		--mmdebug("1")
		--First 6h cycle of the day.
		secondsLeft = secondsToDailyReset - 64800
	elseif (secondsToDailyReset <= 64800 and secondsToDailyReset >= 43200) then
		--mmdebug("2")
		--Second 6h cycle of the day.
		secondsLeft = secondsToDailyReset - 43200
	elseif (secondsToDailyReset <= 43200 and secondsToDailyReset >= 21600) then
		--mmdebug("3")
		--Third 6h cycle of the day.
		secondsLeft = secondsToDailyReset - 21600
	elseif (secondsToDailyReset <= 21600 and secondsToDailyReset >= 1) then
		--mmdebug("4")
		--Fourth 6h cycle of the day.
		secondsLeft = secondsToDailyReset - 1
	end
	Multimod:chatPrefixMsg("Time left until World Quest reset: " .. getTimeString(secondsLeft, true))
end

--Convert both RGB types (0-255) and (0.0-1.0).
function RGBToHex(r, g, b)
	r = tonumber(r)
	g = tonumber(g)
	b = tonumber(b)
	--Check if whole numbers.
	if (r == math.floor(r) and g == math.floor(g) and b == math.floor(b)) then
		r = r <= 255 and r >= 0 and r or 0
		g = g <= 255 and g >= 0 and g or 0
		b = b <= 255 and b >= 0 and b or 0
		return string.format("%02x%02x%02x", r, g, b)
	else
		return string.format("%02x%02x%02x", r*255, g*255, b*255);
	end
end

--Print itemlinks etc in raw format.
function rawPrint(msg)
	local msg = gsub(msg, "\124", "\124\124")
	print(msg)
end

--Show table contents ingame with blizzards built in debug tools.
function var_dump(...)
	UIParentLoadAddOn('Blizzard_DebugTools') -- load debug tools
	DevTools_Dump(...)
    DisplayTableInspectorWindow(...)
end

--Print debug msgs.
function mmdebug(msg, raw)
	if (raw) then
		msg = gsub(msg, "\124", "\124\124")
	end
	print("Debug: " .. msg)
end

function getArmorType()
	local _, class = UnitClass("player");
	local classArray = {
    	["DEATHKNIGHT"] = "PLATE",
    	["DEMONHUNTER"] = "LEATHER",
    	["DRUID"] = "LEATHER",
    	["HUNTER"] = "MAIL",
    	["MAGE"] = "CLOTH",
    	["MONK"] = "LEATHER",
    	["PALADIN"] = "PLATE",
    	["PRIEST"] = "CLOTH",
    	["ROGUE"] = "LEATHER",
    	["SHAMAN"] = "MAIL",
    	["WARLOCK"] = "CLOTH",
    	["WARRIOR"] = "PLATE",
    }
    if (classArray[class]) then
    	return classArray[class]
    else
    	return "Unknown"
    end
end

--Upper case first letter.
function ucfirst(str)
    return (str:gsub("^%l", string.upper))
end